@@ -2,10 +2,12 @@
  * Convenience pkcs11 library that can be linked into an application,
  * and will bind to a specific pkcs11 module.
  *
- * Copyright (C) 2002  Olaf Kirch <okir@lst.de>
+ * Copyright (C) 2002  Olaf Kirch <okir@suse.de>
  */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <stdlib.h>
 #include <stdio.h>
@@ -36,8 +38,10 @@ C_LoadModule(const char *mspec, CK_FUNCTION_LIST_PTR_PTR funcs)
 	mod = calloc(1, sizeof(*mod));
 	mod->_magic = MAGIC;
 
-	if (mspec == NULL)
+	if (mspec == NULL) {
+		free(mod);
 		return NULL;
+	}
 	mod->handle = sc_dlopen(mspec);
 	if (mod->handle == NULL) {
 		fprintf(stderr, "sc_dlopen failed: %s\n", sc_dlerror());
@@ -72,7 +76,7 @@ C_UnloadModule(void *module)
 	if (!mod || mod->_magic != MAGIC)
 		return CKR_ARGUMENTS_BAD;
 
-	if (sc_dlclose(mod->handle) < 0)
+	if (mod->handle != NULL && sc_dlclose(mod->handle) < 0)
 		return CKR_FUNCTION_FAILED;
 
 	memset(mod, 0, sizeof(*mod));
