@@ -18,7 +18,9 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -169,23 +171,6 @@ static int sc_apdu2bytes(sc_context_t *ctx, const sc_apdu_t *apdu,
 	return SC_SUCCESS;
 }
 
-void sc_apdu_log(sc_context_t *ctx, int level, const u8 *data, size_t len, int is_out)
-{
-	size_t blen = len * 5 + 128;
-	char   *buf = malloc(blen);
-	if (buf == NULL)
-		return;
-
-	sc_hex_dump(ctx, level, data, len, buf, blen);
-
-	sc_debug(ctx, level, "\n%s APDU data [%5u bytes] =====================================\n"
-		"%s"
-		"======================================================================\n",
-		is_out != 0 ? "Outgoing" : "Incoming", len,
-		buf);
-	free(buf);
-}
-
 int sc_apdu_get_octets(sc_context_t *ctx, const sc_apdu_t *apdu, u8 **buf,
 	size_t *len, unsigned int proto)
 {
@@ -203,8 +188,10 @@ int sc_apdu_get_octets(sc_context_t *ctx, const sc_apdu_t *apdu, u8 **buf,
 	if (nbuf == NULL)
 		return SC_ERROR_OUT_OF_MEMORY;
 	/* encode the APDU in the buffer */
-	if (sc_apdu2bytes(ctx, apdu, proto, nbuf, nlen) != SC_SUCCESS)
+	if (sc_apdu2bytes(ctx, apdu, proto, nbuf, nlen) != SC_SUCCESS) {
+		free(nbuf);
 		return SC_ERROR_INTERNAL;
+	}
 	*buf = nbuf;
 	*len = nlen;
 
@@ -234,95 +221,6 @@ int sc_apdu_set_resp(sc_context_t *ctx, sc_apdu_t *apdu, const u8 *buf,
 	return SC_SUCCESS;
 }
 
-#ifdef ENABLE_SM
-static const struct sc_asn1_entry c_asn1_sm_response[4] = {
-	{ "encryptedData",	SC_ASN1_OCTET_STRING,   SC_ASN1_CTX | 7,        SC_ASN1_OPTIONAL,       NULL, NULL },
-	{ "statusWord",		SC_ASN1_OCTET_STRING,   SC_ASN1_CTX | 0x19,     0,                      NULL, NULL },
-	{ "mac",		SC_ASN1_OCTET_STRING,   SC_ASN1_CTX | 0x0E,     0,                      NULL, NULL },
-	{ NULL, 0, 0, 0, NULL, NULL }
-};
-static int
-sc_sm_parse_answer(struct sc_context *ctx, unsigned char *resp_data, size_t resp_len,
-		struct sm_card_response *out)
-{
-	struct sc_asn1_entry asn1_sm_response[4];
-	unsigned char data[SC_MAX_APDU_BUFFER_SIZE];
-	size_t data_len = sizeof(data);
-	unsigned char status[2] = {0, 0};
-	size_t status_len = sizeof(status);
-	unsigned char mac[8];
-	size_t mac_len = sizeof(mac);
-	int r;
-
-	if (!resp_data || !resp_len || !out)
-		return SC_ERROR_INVALID_ARGUMENTS;
-
-	sc_copy_asn1_entry(c_asn1_sm_response, asn1_sm_response);
-
-	sc_format_asn1_entry(asn1_sm_response + 0, data, &data_len, 0);
-	sc_format_asn1_entry(asn1_sm_response + 1, status, &status_len, 0);
-	sc_format_asn1_entry(asn1_sm_response + 2, mac, &mac_len, 0);
-
-	r = sc_asn1_decode(ctx, asn1_sm_response, resp_data, resp_len, NULL, NULL);
-	if (r)
-		return r;
-
-	if (asn1_sm_response[1].flags & SC_ASN1_PRESENT)   {
-		out->sw1 = status[0];
-		out->sw2 = status[1];
-	}
-
-	if (asn1_sm_response[2].flags & SC_ASN1_PRESENT)   {
-		memcpy(out->mac, mac, mac_len);
-		out->mac_len = mac_len;
-	}
-	/* TODO: to be continued ... */
-
-	return SC_SUCCESS;
-}
-
-/**  parse answer of SM protected APDU returned by APDU or by 'GET RESPONSE'
- *  @param  card 'sc_card' smartcard object
- *  @param  resp_data 'raw data returned by SM protected APDU
- *  @param  resp_len 'length of raw data returned by SM protected APDU
- *  @param  ref_rv 'status word returned by APDU or 'GET RESPONSE' (can be different from status word encoded into SM response date)
- *  @param  apdu 'sc_apdu' object to update
- *  @return SC_SUCCESS on success and an error code otherwise
- */
-static int
-sc_sm_update_apdu_response(struct sc_card *card, unsigned char *resp_data, size_t resp_len, int ref_rv,
-		struct sc_apdu *apdu)
-{
-	struct sm_card_response sm_resp;
-	int r;
-
-	if (!apdu)
-		return SC_ERROR_INVALID_ARGUMENTS;
-	else if (!resp_data || !resp_len)
-		return SC_SUCCESS;
-
-	memset(&sm_resp, 0, sizeof(sm_resp));
-	r = sc_sm_parse_answer(card->ctx, resp_data, resp_len, &sm_resp);
-	if (r)
-		return r;
-	else if (!sm_resp.sw1 && !sm_resp.sw2)
-		return SC_ERROR_INVALID_DATA;
-	else if (ref_rv != sc_check_sw(card, sm_resp.sw1, sm_resp.sw2))
-		return SC_ERROR_INVALID_DATA;
-
-	if (sm_resp.mac_len)   {
-		if (sm_resp.mac_len > sizeof(apdu->mac))
-			return SC_ERROR_INVALID_DATA;
-		memcpy(apdu->mac, sm_resp.mac, sm_resp.mac_len);
-		apdu->mac_len = sm_resp.mac_len;
-	}
-
-	apdu->sw1 = sm_resp.sw1;
-	apdu->sw2 = sm_resp.sw2;
-
-	return SC_SUCCESS;
-}
-#endif
 
 /*********************************************************************/
 /*   higher level APDU transfer handling functions                   */
@@ -354,12 +252,15 @@ sc_sm_update_apdu_response(struct sc_card *card, unsigned char *resp_data, size_
  *  @param  apdu  sc_apdu_t object to check
  *  @return SC_SUCCESS on success and an error code otherwise
  */
-static int sc_check_apdu(sc_card_t *card, const sc_apdu_t *apdu)
+int
+sc_check_apdu(sc_card_t *card, const sc_apdu_t *apdu)
 {
 	if ((apdu->cse & ~SC_APDU_SHORT_MASK) == 0) {
-		/* length check for short APDU    */
-		if (apdu->le > 256 || (apdu->lc > 255 && (apdu->flags & SC_APDU_FLAGS_CHAINING) == 0))
+		/* length check for short APDU */
+		if (apdu->le > 256 || (apdu->lc > 255 && (apdu->flags & SC_APDU_FLAGS_CHAINING) == 0))   {
+			sc_log(card->ctx, "failed length check for short APDU");
 			goto error;
+		}
 	}
 	else if ((apdu->cse & SC_APDU_EXT) != 0) {
 		/* check if the card supports extended APDUs */
@@ -368,8 +269,10 @@ static int sc_check_apdu(sc_card_t *card, const sc_apdu_t *apdu)
 			goto error;
 		}
 		/* length check for extended APDU */
-		if (apdu->le > 65536 || apdu->lc > 65535)
+		if (apdu->le > 65536 || apdu->lc > 65535)   {
+			sc_log(card->ctx, "failed length check for extended APDU");
 			goto error;
+		}
 	}
 	else   {
 		goto error;
@@ -382,7 +285,7 @@ static int sc_check_apdu(sc_card_t *card, const sc_apdu_t *apdu)
 			goto error;
 		break;
 	case SC_APDU_CASE_2_SHORT:
-		/* no data is sent        */
+		/* no data is sent */
 		if (apdu->datalen != 0 || apdu->lc != 0)
 			goto error;
 		/* data is expected       */
@@ -394,7 +297,7 @@ static int sc_check_apdu(sc_card_t *card, const sc_apdu_t *apdu)
 			goto error;
 		break;
 	case SC_APDU_CASE_3_SHORT:
-		/* data is sent           */
+		/* data is sent */
 		if (apdu->datalen == 0 || apdu->data == NULL || apdu->lc == 0)
 			goto error;
 		/* no data is expected    */
@@ -405,7 +308,7 @@ static int sc_check_apdu(sc_card_t *card, const sc_apdu_t *apdu)
 			goto error;
 		break;
 	case SC_APDU_CASE_4_SHORT:
-		/* data is sent           */
+		/* data is sent */
 		if (apdu->datalen == 0 || apdu->data == NULL || apdu->lc == 0)
 			goto error;
 		/* data is expected       */
@@ -441,7 +344,8 @@ error:
  *  APDU if one of the SC_APDU_CASE_? types is used.
  *  @param  apdu  APDU object
  */
-static void sc_detect_apdu_cse(const sc_card_t *card, sc_apdu_t *apdu)
+static void
+sc_detect_apdu_cse(const sc_card_t *card, sc_apdu_t *apdu)
 {
 	if (apdu->cse == SC_APDU_CASE_2 || apdu->cse == SC_APDU_CASE_3 ||
 	    apdu->cse == SC_APDU_CASE_4) {
@@ -458,48 +362,6 @@ static void sc_detect_apdu_cse(const sc_card_t *card, sc_apdu_t *apdu)
 }
 
 
-#ifdef ENABLE_SM
-static int
-sc_single_sm_transmit(struct sc_card *card, struct sc_apdu *apdu)
-{
-	struct sc_context *ctx  = card->ctx;
-	struct sc_apdu *sm_apdu = NULL;
-	int rv;
-
-	LOG_FUNC_CALLED(ctx);
-	sc_log(ctx, "SM_MODE:%X", card->sm_ctx.sm_mode);
-	if (!card->sm_ctx.ops.get_sm_apdu || !card->sm_ctx.ops.free_sm_apdu)
-		LOG_FUNC_RETURN(ctx, SC_ERROR_NOT_SUPPORTED);
-
-	/* get SM encoded APDU */
-	rv = card->sm_ctx.ops.get_sm_apdu(card, apdu, &sm_apdu);
-	if (rv == SC_ERROR_SM_NOT_APPLIED)   {
-		/* SM wrap of this APDU is ignored by card driver.
-		 * Send plain APDU to the reader driver */
-		rv = card->reader->ops->transmit(card->reader, apdu);
-		LOG_FUNC_RETURN(ctx, rv);
-	}
-	LOG_TEST_RET(ctx, rv, "get SM APDU error");
-
-	/* check if SM APDU is still valid */
-	rv = sc_check_apdu(card, sm_apdu);
-	if (rv < 0)   {
-		card->sm_ctx.ops.free_sm_apdu(card, apdu, &sm_apdu);
-		LOG_TEST_RET(ctx, rv, "cannot validate SM encoded APDU");
-	}
-
-	/* send APDU to the reader driver */
-	rv = card->reader->ops->transmit(card->reader, sm_apdu);
-	LOG_TEST_RET(ctx, rv, "unable to transmit APDU");
-
-	/* decode SM answer and free temporary SM related data */
-	rv = card->sm_ctx.ops.free_sm_apdu(card, apdu, &sm_apdu);
-
-	LOG_FUNC_RETURN(ctx, rv);
-}
-#endif
-
-
 static int
 sc_single_transmit(struct sc_card *card, struct sc_apdu *apdu)
 {
@@ -514,7 +376,7 @@ sc_single_transmit(struct sc_card *card, struct sc_apdu *apdu)
 			apdu->cla, apdu->ins, apdu->p1, apdu->p2, apdu->datalen, apdu->data);
 #ifdef ENABLE_SM
 	if (card->sm_ctx.sm_mode == SM_MODE_TRANSMIT)
-		return sc_single_sm_transmit(card, apdu);
+		return sc_sm_single_transmit(card, apdu);
 #endif
 
 	/* send APDU to the reader driver */
@@ -524,6 +386,7 @@ sc_single_transmit(struct sc_card *card, struct sc_apdu *apdu)
 	LOG_FUNC_RETURN(ctx, rv);
 }
 
+
 static int
 sc_set_le_and_transmit(struct sc_card *card, struct sc_apdu *apdu, size_t olen)
 {
@@ -628,7 +491,7 @@ sc_get_response(struct sc_card *card, struct sc_apdu *apdu, size_t olen)
 			/* if the card has returned 0x9000 but we still expect data ask for more
 			 * until we have read enough bytes */
 			le = minlen;
-	} while (rv != 0 || minlen != 0);
+	} while (rv != 0 && minlen != 0);
 
 	/* we've read all data, let's return 0x9000 */
 	apdu->resplen = buf - apdu->resp;
@@ -705,7 +568,7 @@ int sc_transmit_apdu(sc_card_t *card, sc_apdu_t *apdu)
 		 * bytes using command chaining */
 		size_t    len  = apdu->datalen;
 		const u8  *buf = apdu->data;
-		size_t    max_send_size = card->max_send_size > 0 ? card->max_send_size : 255;
+		size_t    max_send_size = sc_get_max_send_size(card);
 
 		while (len != 0) {
 			size_t    plen;
