@@ -18,7 +18,9 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <string.h>
 
@@ -26,6 +28,8 @@
 #include "cardctl.h"
 
 static struct sc_atr_table acos5_atrs[] = {
+	{"3b:be:96:00:00:41:05:20:00:00:00:00:00:00:00:00:00:90:00", NULL, NULL,
+	SC_CARD_TYPE_ACOS5_GENERIC, 0, NULL},
 	{"3b:be:18:00:00:41:05:10:00:00:00:00:00:00:00:00:00:90:00", NULL, NULL,
 	 SC_CARD_TYPE_ACOS5_GENERIC, 0, NULL},
 	{NULL, NULL, NULL, 0, 0, NULL}
