@@ -19,7 +19,9 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <sys/types.h>
 #include <assert.h>
@@ -112,7 +114,7 @@ static int token_init(sc_card_t *card, const char *card_name)
 	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_NORMAL);
 
 	card->name = card_name;
-	card->caps |= SC_CARD_CAP_NO_FCI | SC_CARD_CAP_RNG;
+	card->caps |= SC_CARD_CAP_RNG;
 	card->drv_data = calloc(1, sizeof(auth_senv_t));
 	if (card->drv_data == NULL)
 		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_OUT_OF_MEMORY);
@@ -1113,7 +1115,7 @@ static int rutoken_get_challenge(sc_card_t *card, u8 *rnd, size_t count)
 	sc_apdu_t apdu;
 	u8 rbuf[32];
 	size_t n;
-	int ret = SC_ERROR_INVALID_ARGUMENTS; /* if count == 0 */
+	int ret = SC_SUCCESS; /* if count == 0 */
 
 	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0x84, 0x00, 0x00);
