@@ -1,7 +1,7 @@
 /*
  * card-tcos.c: Support for TCOS cards
  *
- * Copyright (C) 2007  Peter Koch <Koch@smartcard-auth.de>
+ * Copyright (C) 2011  Peter Koch <pk@opensc-project.org>
  * Copyright (C) 2002  g10 Code GmbH
  * Copyright (C) 2001  Juha Yrjölä <juha.yrjola@iki.fi>
  *
@@ -20,7 +20,9 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <string.h>
 #include <ctype.h>
@@ -42,6 +44,8 @@ static struct sc_atr_table tcos_atrs[] = {
 	{ "3B:BA:96:00:81:31:86:5D:00:64:05:7B:02:03:31:80:90:00:7D", NULL, NULL, SC_CARD_TYPE_TCOS_V2, 0, NULL },
 	/* Philips P5CT072 */
 	{ "3B:BF:96:00:81:31:FE:5D:00:64:04:11:03:01:31:C0:73:F7:01:D0:00:90:00:7D", NULL, NULL, SC_CARD_TYPE_TCOS_V3, 0, NULL },
+	/* Philips P5CT080 */
+	{ "3B:BF:B6:00:81:31:FE:5D:00:64:04:28:03:02:31:C0:73:F7:01:D0:00:90:00:67", NULL, NULL, SC_CARD_TYPE_TCOS_V3, 0, NULL },
 	{ NULL, NULL, NULL, 0, 0, NULL }
 };
 
@@ -159,8 +163,6 @@ static int tcos_construct_fci(const sc_file_t *file,
         /* Directory name */
         if (file->type == SC_FILE_TYPE_DF) {
                 if (file->namelen) {
-                        if (file->namelen > 16 || !file->name)
-                                return SC_ERROR_INVALID_ARGUMENTS;
                         sc_asn1_put_tag(0x84, file->name, file->namelen,
                                         p, 16, &p);
                 }
@@ -353,6 +355,7 @@ static int tcos_select_file(sc_card_t *card,
 	switch (in_path->type) {
 	case SC_PATH_TYPE_FILE_ID:
 		if (pathlen != 2) return SC_ERROR_INVALID_ARGUMENTS;
+		/* fall through */
 	case SC_PATH_TYPE_FROM_CURRENT:
 		apdu.p1 = 9;
 		break;
@@ -400,6 +403,7 @@ static int tcos_select_file(sc_card_t *card,
 
 	file = sc_file_new();
 	if (file == NULL) SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_OUT_OF_MEMORY);
+	*file_out = file;
 	file->path = *in_path;
 
 	for(i=2; i+1<apdu.resplen && i+1+apdu.resp[i+1]<apdu.resplen; i+=2+apdu.resp[i+1]){
@@ -438,7 +442,6 @@ static int tcos_select_file(sc_card_t *card,
 		}
 	}
 	file->magic = SC_FILE_MAGIC;
-	*file_out = file;
 
 	parse_sec_attr(card, file, file->sec_attr, file->sec_attr_len);
 
