@@ -18,7 +18,9 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <assert.h>
 #include <string.h>
@@ -215,7 +217,7 @@ static int westcos_init(sc_card_t * card)
 	card->drv_data = malloc(sizeof(priv_data_t));
 	if (card->drv_data == NULL)
 		return SC_ERROR_OUT_OF_MEMORY;
-	memset(card->drv_data, 0, sizeof(card->drv_data));
+	memset(card->drv_data, 0, sizeof(priv_data_t));
 	
 	priv_data = (priv_data_t *) card->drv_data;
 
@@ -533,6 +535,7 @@ static int westcos_create_file(sc_card_t *card, struct sc_file *file)
 		break;
 	case SC_FILE_TYPE_INTERNAL_EF:
 		buf[0] |= 0x80;
+		/* fall through */
 	case SC_FILE_TYPE_WORKING_EF:
 		switch (file->ef_structure) {
 		case SC_FILE_EF_TRANSPARENT:
@@ -1095,12 +1098,12 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 				 size_t outlen)
 {
 	int r;
+	sc_file_t *keyfile = NULL;
+#ifdef ENABLE_OPENSSL
 	int idx = 0;
 	u8 buf[180];
-	sc_file_t *keyfile = sc_file_new();
 	priv_data_t *priv_data = NULL;
 	int pad;
-#ifdef ENABLE_OPENSSL
 	RSA *rsa = NULL;
 	BIO *mem = BIO_new(BIO_s_mem());
 #endif
@@ -1109,6 +1112,15 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 		return SC_ERROR_INVALID_ARGUMENTS;
 	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
 		"westcos_sign_decipher outlen=%d\n", outlen);
+	
+#ifndef ENABLE_OPENSSL
+	r = SC_ERROR_NOT_SUPPORTED;
+#else
+	if (mem == NULL || card->drv_data == NULL) {
+		r = SC_ERROR_OUT_OF_MEMORY;
+		goto out;
+	}
+
 	priv_data = (priv_data_t *) card->drv_data;
 
 	if(priv_data->flags & RSA_CRYPTO_COMPONENT)
@@ -1134,14 +1146,6 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 		r = apdu.resplen;
 		goto out2;
 	}
-	
-#ifndef ENABLE_OPENSSL
-	r = SC_ERROR_NOT_SUPPORTED;
-#else
-	if (keyfile == NULL || mem == NULL || priv_data == NULL) {
-		r = SC_ERROR_OUT_OF_MEMORY;
-		goto out;
-	}
 	if ((priv_data->env.flags) & SC_ALGORITHM_RSA_PAD_PKCS1)
 		pad = RSA_PKCS1_PADDING;
 
@@ -1153,7 +1157,7 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 		goto out;
 	}
 	r = sc_select_file(card, &(priv_data->env.file_ref), &keyfile);
-	if (r)
+	if (r || !keyfile)
 		goto out;
 
 	do {
@@ -1231,8 +1235,8 @@ out:
 		BIO_free(mem);
 	if (rsa)
 		RSA_free(rsa);
-#endif /* ENABLE_OPENSSL */
 out2:
+#endif /* ENABLE_OPENSSL */
 	if (keyfile)
 		sc_file_free(keyfile);
 	return r;
