@@ -18,7 +18,9 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <assert.h>
 #include <stdlib.h>
@@ -29,6 +31,7 @@
 
 #include "internal.h"
 #include "asn1.h"
+#include "common/compat_strlcpy.h"
 
 /*
 #define INVALIDATE_CARD_CACHE_IN_UNLOCK
@@ -94,15 +97,35 @@ static void sc_card_free(sc_card_t *card)
 {
 	sc_free_apps(card);
 	sc_free_ef_atr(card);
+
 	if (card->ef_dir != NULL)
 		sc_file_free(card->ef_dir);
+
 	free(card->ops);
-	if (card->algorithms != NULL)
+
+	if (card->algorithms != NULL)   {
+		int i;
+		for (i=0; i<card->algorithm_count; i++)   {
+			struct sc_algorithm_info *info = (card->algorithms + i);
+			if (info->algorithm == SC_ALGORITHM_EC)   {
+				struct sc_ec_parameters ep = info->u._ec.params;
+
+				free(ep.named_curve);
+				free(ep.der.value);
+			}
+		}
 		free(card->algorithms);
+
+		card->algorithms = NULL;
+		card->algorithm_count = 0;
+	}
+
 	if (card->cache.current_ef)
 		sc_file_free(card->cache.current_ef);
+
 	if (card->cache.current_df)
 		sc_file_free(card->cache.current_df);
+
 	if (card->mutex != NULL) {
 		int r = sc_mutex_destroy(card->ctx, card->mutex);
 		if (r != SC_SUCCESS)
@@ -112,6 +135,54 @@ static void sc_card_free(sc_card_t *card)
 	free(card);
 }
 
+size_t sc_get_max_recv_size(const sc_card_t *card)
+{
+	size_t max_recv_size;
+	assert(card != NULL && card->reader != NULL);
+	max_recv_size = card->max_recv_size;
+
+	/* initialize max_recv_size to a meaningfull value */
+	if (card->caps & SC_CARD_CAP_APDU_EXT) {
+		if (!max_recv_size)
+			max_recv_size = 65536;
+	} else {
+		if (!max_recv_size)
+			max_recv_size = 256;
+	}
+
+	/*  Override card limitations with reader limitations. */
+	if (card->reader->max_recv_size != 0
+			&& (card->reader->max_recv_size < card->max_recv_size))
+		max_recv_size = card->reader->max_recv_size;
+
+	return max_recv_size;
+}
+
+size_t sc_get_max_send_size(const sc_card_t *card)
+{
+	size_t max_send_size;
+
+	assert(card != NULL && card->reader != NULL);
+
+	max_send_size = card->max_send_size;
+
+	/* initialize max_send_size to a meaningfull value */
+	if (card->caps & SC_CARD_CAP_APDU_EXT) {
+		if (!max_send_size)
+			max_send_size = 65535;
+	} else {
+		if (!max_send_size)
+			max_send_size = 255;
+	}
+
+	/*  Override card limitations with reader limitations. */
+	if (card->reader->max_send_size != 0
+			&& (card->reader->max_send_size < card->max_send_size))
+		max_send_size = card->reader->max_send_size;
+
+	return max_send_size;
+}
+
 int sc_connect_card(sc_reader_t *reader, sc_card_t **card_out)
 {
 	sc_card_t *card;
@@ -169,10 +240,14 @@ int sc_connect_card(sc_reader_t *reader, sc_card_t **card_out)
 	}
 
 	if (driver != NULL) {
-		/* Forced driver, or matched via ATR mapping from
-		 * config file */
+		/* Forced driver, or matched via ATR mapping from config file */
 		card->driver = driver;
+
 		memcpy(card->ops, card->driver->ops, sizeof(struct sc_card_operations));
+		if (card->ops->match_card != NULL)
+			if (card->ops->match_card(card) != 1)
+				sc_log(ctx, "driver '%s' match_card() failed: %s (will continue anyway)", card->driver->name, sc_strerror(r));
+
 		if (card->ops->init != NULL) {
 			r = card->ops->init(card);
 			if (r) {
@@ -180,15 +255,23 @@ int sc_connect_card(sc_reader_t *reader, sc_card_t **card_out)
 				goto err;
 			}
 		}
-	} else {
-		sc_debug(ctx, SC_LOG_DEBUG_MATCH, "matching built-in ATRs");
+	}
+	else {
+		sc_log(ctx, "matching built-in ATRs");
 		for (i = 0; ctx->card_drivers[i] != NULL; i++) {
 			struct sc_card_driver *drv = ctx->card_drivers[i];
 			const struct sc_card_operations *ops = drv->ops;
 
 			sc_log(ctx, "trying driver '%s'", drv->short_name);
-			if (ops == NULL || ops->match_card == NULL)
+			if (ops == NULL || ops->match_card == NULL)   {
 				continue;
+			}
+			else if (!(ctx->flags & SC_CTX_FLAG_ENABLE_DEFAULT_DRIVER)
+				   	&& !strcmp("default", drv->short_name))   {
+				sc_log(ctx , "ignore 'default' card driver");
+				continue;
+			}
+
 			/* Needed if match_card() needs to talk with the card (e.g. card-muscle) */
 			*card->ops = *ops;
 			if (ops->match_card(card) != 1)
@@ -215,18 +298,10 @@ int sc_connect_card(sc_reader_t *reader, sc_card_t **card_out)
 	}
 	if (card->name == NULL)
 		card->name = card->driver->name;
-	*card_out = card;
 
-        /*  Override card limitations with reader limitations.
-         *  Note that zero means no limitations at all.
-	 */
-        if ((card->max_recv_size == 0) ||
-           ((reader->driver->max_recv_size != 0) && (reader->driver->max_recv_size < card->max_recv_size)))
-                card->max_recv_size = reader->driver->max_recv_size;
-
-        if ((card->max_send_size == 0) ||
-           ((reader->driver->max_send_size != 0) && (reader->driver->max_send_size < card->max_send_size)))
-                card->max_send_size = reader->driver->max_send_size;
+	/* initialize max_send_size/max_recv_size to a meaningfull value */
+	card->max_recv_size = sc_get_max_recv_size(card);
+	card->max_send_size = sc_get_max_send_size(card);
 
 	sc_log(ctx, "card info name:'%s', type:%i, flags:0x%X, max_send/recv_size:%i/%i",
 		card->name, card->type, card->flags, card->max_send_size, card->max_recv_size);
@@ -239,6 +314,7 @@ int sc_connect_card(sc_reader_t *reader, sc_card_t **card_out)
 		goto err;
 	}
 #endif
+	*card_out = card;
 
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 err:
@@ -312,10 +388,11 @@ int sc_lock(sc_card_t *card)
 {
 	int r = 0, r2 = 0;
 
-	LOG_FUNC_CALLED(card->ctx);
-
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
+
+	LOG_FUNC_CALLED(card->ctx);
+
 	r = sc_mutex_lock(card->ctx, card->mutex);
 	if (r != SC_SUCCESS)
 		return r;
@@ -326,6 +403,10 @@ int sc_lock(sc_card_t *card)
 				/* invalidate cache */
 				memset(&card->cache, 0, sizeof(card->cache));
 				card->cache.valid = 0;
+#ifdef ENABLE_SM
+				if (card->sm_ctx.ops.open)
+					card->sm_ctx.ops.open(card);
+#endif
 				r = card->reader->ops->lock(card->reader);
 			}
 		}
@@ -395,15 +476,16 @@ int sc_create_file(sc_card_t *card, sc_file_t *file)
 {
 	int r;
 	char pbuf[SC_MAX_PATH_STRING_SIZE];
-	const sc_path_t *in_path = &file->path;
+	const sc_path_t *in_path;
 
-	assert(card != NULL);
+	assert(card != NULL && file != NULL);
 
+	in_path = &file->path;
 	r = sc_path_print(pbuf, sizeof(pbuf), in_path);
 	if (r != SC_SUCCESS)
 		pbuf[0] = '\0';
 
-	sc_log(card->ctx, "called; type=%d, path=%s, size=%u",  in_path->type, pbuf, file->size);
+	sc_log(card->ctx, "called; type=%d, path=%s, id=%04i, size=%u",  in_path->type, pbuf, file->id, file->size);
 	/* ISO 7816-4: "Number of data bytes in the file, including structural information if any"
 	 * can not be bigger than two bytes */
 	if (file->size > 0xFFFF)
@@ -438,7 +520,7 @@ int sc_delete_file(sc_card_t *card, const sc_path_t *path)
 int sc_read_binary(sc_card_t *card, unsigned int idx,
 		   unsigned char *buf, size_t count, unsigned long flags)
 {
-	size_t max_le = card->max_recv_size > 0 ? card->max_recv_size : 256;
+	size_t max_le = sc_get_max_recv_size(card);
 	int r;
 
 	assert(card != NULL && card->ops != NULL && buf != NULL);
@@ -488,7 +570,7 @@ int sc_read_binary(sc_card_t *card, unsigned int idx,
 int sc_write_binary(sc_card_t *card, unsigned int idx,
 		    const u8 *buf, size_t count, unsigned long flags)
 {
-	size_t max_lc = card->max_send_size > 0 ? card->max_send_size : 255;
+	size_t max_lc = sc_get_max_send_size(card);
 	int r;
 
 	assert(card != NULL && card->ops != NULL && buf != NULL);
@@ -531,7 +613,7 @@ int sc_write_binary(sc_card_t *card, unsigned int idx,
 int sc_update_binary(sc_card_t *card, unsigned int idx,
 		     const u8 *buf, size_t count, unsigned long flags)
 {
-	size_t max_lc = card->max_send_size > 0 ? card->max_send_size : 255;
+	size_t max_lc = sc_get_max_send_size(card);
 	int r;
 
 	assert(card != NULL && card->ops != NULL && buf != NULL);
@@ -607,6 +689,12 @@ int sc_select_file(sc_card_t *card, const sc_path_t *in_path,  sc_file_t **file)
 	if (r != SC_SUCCESS)
 		pbuf[0] = '\0';
 
+	/* FIXME We should be a bit less strict and let the upper layers do
+	 * the initialization (including reuse of existing file objects). We
+	 * implemented this here because we are lazy. */
+	if (file != NULL)
+		*file = NULL;
+
 	sc_log(card->ctx, "called; type=%d, path=%s", in_path->type, pbuf);
 	if (in_path->len > SC_MAX_PATH_SIZE)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
@@ -631,9 +719,16 @@ int sc_select_file(sc_card_t *card, const sc_path_t *in_path,  sc_file_t **file)
 	r = card->ops->select_file(card, in_path, file);
 	LOG_TEST_RET(card->ctx, r, "'SELECT' error");
 
-	/* Remember file path */
-	if (file && *file)
-		(*file)->path = *in_path;
+	if (file) {
+		if (*file)
+			/* Remember file path */
+			(*file)->path = *in_path;
+		else
+			/* FIXME We should be a bit less strict and let the upper layers do
+			 * the error checking. We implemented this here because we are
+			 * lazy.  */
+			r = SC_ERROR_INVALID_DATA;
+	}
 
 	LOG_FUNC_RETURN(card->ctx, r);
 }
@@ -792,21 +887,27 @@ int _sc_card_add_algorithm(sc_card_t *card, const sc_algorithm_info_t *info)
 }
 
 int  _sc_card_add_ec_alg(sc_card_t *card, unsigned int key_length,
-			unsigned long flags, unsigned long ext_flags)
+			unsigned long flags, unsigned long ext_flags,
+			struct sc_object_id *curve_oid)
 {
 	sc_algorithm_info_t info;
 
 	memset(&info, 0, sizeof(info));
+	sc_init_oid(&info.u._ec.params.id);
+
 	info.algorithm = SC_ALGORITHM_EC;
 	info.key_length = key_length;
 	info.flags = flags;
+
 	info.u._ec.ext_flags = ext_flags;
+	if (curve_oid)
+		info.u._ec.params.id = *curve_oid;
 
 	return _sc_card_add_algorithm(card, &info);
 }
 
 static sc_algorithm_info_t * sc_card_find_alg(sc_card_t *card,
-		unsigned int algorithm, unsigned int key_length)
+		unsigned int algorithm, unsigned int key_length, void *param)
 {
 	int i;
 
@@ -817,15 +918,20 @@ static sc_algorithm_info_t * sc_card_find_alg(sc_card_t *card,
 			continue;
 		if (info->key_length != key_length)
 			continue;
+		if (param)   {
+			if (info->algorithm == SC_ALGORITHM_EC)
+				if(!sc_compare_oid((struct sc_object_id *)param, &info->u._ec.params.id))
+					continue;
+		}
 		return info;
 	}
 	return NULL;
 }
 
 sc_algorithm_info_t * sc_card_find_ec_alg(sc_card_t *card,
-		unsigned int key_length)
+		unsigned int key_length, struct sc_object_id *curve_name)
 {
-	return sc_card_find_alg(card, SC_ALGORITHM_EC, key_length);
+	return sc_card_find_alg(card, SC_ALGORITHM_EC, key_length, curve_name);
 }
 
 int _sc_card_add_rsa_alg(sc_card_t *card, unsigned int key_length,
@@ -845,25 +951,27 @@ int _sc_card_add_rsa_alg(sc_card_t *card, unsigned int key_length,
 sc_algorithm_info_t * sc_card_find_rsa_alg(sc_card_t *card,
 		unsigned int key_length)
 {
-	return sc_card_find_alg(card, SC_ALGORITHM_RSA, key_length);
+	return sc_card_find_alg(card, SC_ALGORITHM_RSA, key_length, NULL);
 }
 
 sc_algorithm_info_t * sc_card_find_gostr3410_alg(sc_card_t *card,
 		unsigned int key_length)
 {
-	return sc_card_find_alg(card, SC_ALGORITHM_GOSTR3410, key_length);
+	return sc_card_find_alg(card, SC_ALGORITHM_GOSTR3410, key_length, NULL);
 }
 
 static int match_atr_table(sc_context_t *ctx, struct sc_atr_table *table, struct sc_atr *atr)
 {
-	u8 *card_atr_bin = atr->value;
-	size_t card_atr_bin_len = atr->len;
+	u8 *card_atr_bin;
+	size_t card_atr_bin_len;
 	char card_atr_hex[3 * SC_MAX_ATR_SIZE];
 	size_t card_atr_hex_len;
 	unsigned int i = 0;
 
 	if (ctx == NULL || table == NULL || atr == NULL)
 		return -1;
+	card_atr_bin = atr->value;
+	card_atr_bin_len = atr->len;
 	sc_bin_to_hex(card_atr_bin, card_atr_bin_len, card_atr_hex, sizeof(card_atr_hex), ':');
 	card_atr_hex_len = strlen(card_atr_hex);
 
@@ -968,32 +1076,40 @@ int _sc_add_atr(sc_context_t *ctx, struct sc_card_driver *driver, struct sc_atr_
 	if (!map)
 		return SC_ERROR_OUT_OF_MEMORY;
 	driver->atr_map = map;
+
 	dst = &driver->atr_map[driver->natrs++];
 	memset(dst, 0, sizeof(*dst));
 	memset(&driver->atr_map[driver->natrs], 0, sizeof(struct sc_atr_table));
 	dst->atr = strdup(src->atr);
 	if (!dst->atr)
 		return SC_ERROR_OUT_OF_MEMORY;
+
 	if (src->atrmask) {
 		dst->atrmask = strdup(src->atrmask);
 		if (!dst->atrmask)
 			return SC_ERROR_OUT_OF_MEMORY;
-	} else {
+	}
+	else {
 		dst->atrmask = NULL;
 	}
+
 	if (src->name) {
 		dst->name = strdup(src->name);
 		if (!dst->name)
 			return SC_ERROR_OUT_OF_MEMORY;
-	} else {
+	}
+	else {
 		dst->name = NULL;
 	}
+
 	dst->type = src->type;
 	dst->flags = src->flags;
 	dst->card_atr = src->card_atr;
+
 	return SC_SUCCESS;
 }
 
+
 int _sc_free_atr(sc_context_t *ctx, struct sc_card_driver *driver)
 {
 	unsigned int i;
@@ -1059,6 +1175,36 @@ void sc_print_cache(struct sc_card *card)   {
 				sc_print_path(&card->cache.current_df->path));
 }
 
+int sc_copy_ec_params(struct sc_ec_parameters *dst, struct sc_ec_parameters *src)
+{
+	if (!dst || !src)
+		return SC_ERROR_INVALID_ARGUMENTS;
+
+	memset(dst, 0, sizeof(*dst));
+	if (src->named_curve)   {
+		dst->named_curve = strdup(src->named_curve);
+		if (!dst->named_curve)
+			return SC_ERROR_OUT_OF_MEMORY;
+	}
+	dst->id = src->id;
+	if (src->der.value && src->der.len)   {
+		dst->der.value = malloc(src->der.len);
+		if (!dst->der.value)
+			return SC_ERROR_OUT_OF_MEMORY;
+		memcpy(dst->der.value, src->der.value, src->der.len);
+		dst->der.len = src->der.len;
+	}
+	src->type = dst->type;
+	src->field_length = dst->field_length;
+
+	return SC_SUCCESS;
+}
+
+scconf_block *
+sc_match_atr_block(sc_context_t *ctx, struct sc_card_driver *driver, struct sc_atr *atr)
+{
+	return _sc_match_atr_block(ctx, driver, atr);
+}
 
 #ifdef ENABLE_SM
 static int
@@ -1083,8 +1229,6 @@ sc_card_sm_load(struct sc_card *card, const char *module_path, const char *in_mo
 #ifdef _WIN32
 	char temp_path[PATH_MAX];
 	int temp_len;
-	long rc;
-	HKEY hKey;
 	const char path_delim = '\\';
 #else
 	const char path_delim = '/';
@@ -1097,25 +1241,12 @@ sc_card_sm_load(struct sc_card *card, const char *module_path, const char *in_mo
 		return sc_card_sm_unload(card);
 
 #ifdef _WIN32
-	if (!module_path) {
-		rc = RegOpenKeyEx( HKEY_CURRENT_USER, "Software\\OpenSC Project\\OpenSC", 0, KEY_QUERY_VALUE, &hKey );
-		if( rc == ERROR_SUCCESS ) {
-			temp_len = PATH_MAX;
-			rc = RegQueryValueEx( hKey, "SmDir", NULL, NULL, (LPBYTE) temp_path, &temp_len);
-			if( (rc == ERROR_SUCCESS) && (temp_len < PATH_MAX) )
-				module_path = temp_path;
-			RegCloseKey( hKey );
-		}
-	}
-	if (!module_path) {
-		rc = RegOpenKeyEx( HKEY_LOCAL_MACHINE, "Software\\OpenSC Project\\OpenSC", 0, KEY_QUERY_VALUE, &hKey );
-		if( rc == ERROR_SUCCESS ) {
-			temp_len = PATH_MAX;
-			rc = RegQueryValueEx( hKey, "SmDir", NULL, NULL, (LPBYTE) temp_path, &temp_len);
-			if(rc == ERROR_SUCCESS && temp_len < PATH_MAX)
-				module_path = temp_path;
-			RegCloseKey( hKey );
-		}
+	if (!module_path)   {
+		temp_len = PATH_MAX;
+		rv = sc_ctx_win32_get_config_value(NULL, "SmDir", "Software\\OpenSC Project\\OpenSC",
+				temp_path, &temp_len);
+		if (rv == SC_SUCCESS)
+			module_path = temp_path;
 	}
 #endif
 	sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "SM module '%s' located in '%s'", in_module, module_path);
@@ -1130,7 +1261,7 @@ sc_card_sm_load(struct sc_card *card, const char *module_path, const char *in_mo
 	}
 
 	if (!module)
-		return SC_ERROR_MEMORY_FAILURE;
+		return SC_ERROR_OUT_OF_MEMORY;
 
 	sc_log(ctx, "try to load SM module '%s'", module);
 	do  {
@@ -1197,6 +1328,7 @@ sc_card_sm_check(struct sc_card *card)
 	int rv, ii;
 
 	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_NORMAL);
+	sc_log(ctx, "card->sm_ctx.ops.open %p", card->sm_ctx.ops.open);
 
 	/* get the name of card specific SM configuration section */
 	atrblock = _sc_match_atr_block(ctx, card->driver, &card->atr);
@@ -1233,8 +1365,8 @@ sc_card_sm_check(struct sc_card *card)
 	rv = sc_card_sm_load(card, module_path, module_name);
 	LOG_TEST_RET(ctx, rv, "Failed to load SM module");
 
-	strncpy(card->sm_ctx.module.filename, module_name, sizeof(card->sm_ctx.module.filename));
-	strncpy(card->sm_ctx.config_section, sm, sizeof(card->sm_ctx.config_section));
+	strlcpy(card->sm_ctx.module.filename, module_name, sizeof(card->sm_ctx.module.filename));
+	strlcpy(card->sm_ctx.config_section, sm, sizeof(card->sm_ctx.config_section));
 
 	/* allocate resources for the external SM module */
 	sc_log(ctx, "'module_init' handler %p", card->sm_ctx.module.ops.module_init);
