@@ -20,14 +20,16 @@
 #ifndef COMPRESSION_H
 #define COMPRESSION_H
 
-#include "opensc.h"
-#include "types.h"
+#include "libopensc/opensc.h"
+#include "libopensc/types.h"
 
 #define COMPRESSION_AUTO	0
 #define COMPRESSION_ZLIB	1
 #define COMPRESSION_GZIP	2
 #define COMPRESSION_UNKNOWN (-1)
 
+int sc_compress(u8* out, size_t* outLen, const u8* in, size_t inLen, int method);
+
 int sc_decompress_alloc(u8** out, size_t* outLen, const u8* in, size_t inLen, int method);
 int sc_decompress(u8* out, size_t* outLen, const u8* in, size_t inLen, int method);
 
