@@ -29,15 +29,34 @@
 #include "asn1.h"
 #include "iso7816.h"
 
+
+static void fixup_transceive_length(const struct sc_card *card,
+		struct sc_apdu *apdu)
+{
+	assert(card != NULL && apdu != NULL);
+
+	if (apdu->lc > sc_get_max_send_size(card)) {
+		/* The lower layers will automatically do chaining */
+		apdu->flags |= SC_APDU_FLAGS_CHAINING;
+	}
+
+	if (apdu->le > sc_get_max_recv_size(card)) {
+		/* The lower layers will automatically do a GET RESPONSE, if possible.
+		 * All other workarounds must be carried out by the upper layers. */
+		apdu->le = sc_get_max_recv_size(card);
+	}
+}
+
+
 static const struct sc_card_error iso7816_errors[] = {
-	{ 0x6200, SC_ERROR_MEMORY_FAILURE,	"State of non-volatile memory unchanged" },
+	{ 0x6200, SC_ERROR_CARD_CMD_FAILED,	"Warning: no information given, non-volatile memory is unchanged" },
 	{ 0x6281, SC_ERROR_CORRUPTED_DATA,	"Part of returned data may be corrupted" },
 	{ 0x6282, SC_ERROR_FILE_END_REACHED,	"End of file/record reached before reading Le bytes" },
 	{ 0x6283, SC_ERROR_CARD_CMD_FAILED,	"Selected file invalidated" },
 	{ 0x6284, SC_ERROR_CARD_CMD_FAILED,	"FCI not formatted according to ISO 7816-4" },
 
-	{ 0x6300, SC_ERROR_MEMORY_FAILURE,	"State of non-volatile memory changed" },
-	{ 0x6381, SC_ERROR_CARD_CMD_FAILED,	"File filled up by last write" },
+	{ 0x6300, SC_ERROR_CARD_CMD_FAILED,	"Warning: no information given, non-volatile memory has changed" },
+	{ 0x6381, SC_ERROR_CARD_CMD_FAILED,	"Warning: file filled up by last write" },
 
 	{ 0x6581, SC_ERROR_MEMORY_FAILURE,	"Memory failure" },
 
@@ -67,246 +86,243 @@ static const struct sc_card_error iso7816_errors[] = {
 	{ 0x6A86, SC_ERROR_INCORRECT_PARAMETERS,"Incorrect parameters P1-P2" },
 	{ 0x6A87, SC_ERROR_INCORRECT_PARAMETERS,"Lc inconsistent with P1-P2" },
 	{ 0x6A88, SC_ERROR_DATA_OBJECT_NOT_FOUND,"Referenced data not found" },
+	{ 0x6A89, SC_ERROR_FILE_ALREADY_EXISTS,  "File already exists"},
+	{ 0x6A8A, SC_ERROR_FILE_ALREADY_EXISTS,  "DF name already exists"},
 
 	{ 0x6B00, SC_ERROR_INCORRECT_PARAMETERS,"Wrong parameter(s) P1-P2" },
 	{ 0x6D00, SC_ERROR_INS_NOT_SUPPORTED,	"Instruction code not supported or invalid" },
 	{ 0x6E00, SC_ERROR_CLASS_NOT_SUPPORTED,	"Class not supported" },
 	{ 0x6F00, SC_ERROR_CARD_CMD_FAILED,	"No precise diagnosis" },
 
-	/* Possibly TCOS / Micardo specific errors */
-	{ 0x6600, SC_ERROR_INCORRECT_PARAMETERS, "Error setting the security env"},
-	{ 0x66F0, SC_ERROR_INCORRECT_PARAMETERS, "No space left for padding"},
-	{ 0x69F0, SC_ERROR_NOT_ALLOWED,          "Command not allowed"},
-	{ 0x6A89, SC_ERROR_FILE_ALREADY_EXISTS,  "Files exists"},
-	{ 0x6A8A, SC_ERROR_FILE_ALREADY_EXISTS,  "Application exists"},
+
+
 };
 
-static int iso7816_check_sw(sc_card_t *card, unsigned int sw1, unsigned int sw2)
+
+static int
+iso7816_check_sw(struct sc_card *card, unsigned int sw1, unsigned int sw2)
 {
 	const int err_count = sizeof(iso7816_errors)/sizeof(iso7816_errors[0]);
 	int i;
-	
+
 	/* Handle special cases here */
 	if (sw1 == 0x6C) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Wrong length; correct length is %d", sw2);
+		sc_log(card->ctx, "Wrong length; correct length is %d", sw2);
 		return SC_ERROR_WRONG_LENGTH;
 	}
 	if (sw1 == 0x90)
 		return SC_SUCCESS;
         if (sw1 == 0x63U && (sw2 & ~0x0fU) == 0xc0U ) {
-             sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Verification failed (remaining tries: %d)",
-                   (sw2 & 0x0f));
+             sc_log(card->ctx, "Verification failed (remaining tries: %d)", (sw2 & 0x0f));
              return SC_ERROR_PIN_CODE_INCORRECT;
         }
-	for (i = 0; i < err_count; i++)
+	for (i = 0; i < err_count; i++)   {
 		if (iso7816_errors[i].SWs == ((sw1 << 8) | sw2)) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "%s", iso7816_errors[i].errorstr);
+			sc_log(card->ctx, "%s", iso7816_errors[i].errorstr);
 			return iso7816_errors[i].errorno;
 		}
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Unknown SWs; SW1=%02X, SW2=%02X", sw1, sw2);
+	}
+
+	sc_log(card->ctx, "Unknown SWs; SW1=%02X, SW2=%02X", sw1, sw2);
 	return SC_ERROR_CARD_CMD_FAILED;
 }
 
-static int iso7816_read_binary(sc_card_t *card,
-			       unsigned int idx, u8 *buf, size_t count,
-			       unsigned long flags)
+
+static int
+iso7816_read_binary(struct sc_card *card, unsigned int idx, u8 *buf, size_t count, unsigned long flags)
 {
-	sc_context_t *ctx = card->ctx;
-	sc_apdu_t apdu;
-	u8 recvbuf[SC_MAX_APDU_BUFFER_SIZE];
+	struct sc_context *ctx = card->ctx;
+	struct sc_apdu apdu;
 	int r;
 
 	if (idx > 0x7fff) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "invalid EF offset: 0x%X > 0x7FFF", idx);
+		sc_log(ctx, "invalid EF offset: 0x%X > 0x7FFF", idx);
 		return SC_ERROR_OFFSET_TOO_LARGE;
 	}
 
-	assert(count <= (card->max_recv_size > 0 ? card->max_recv_size : 256));
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xB0, (idx >> 8) & 0x7F, idx & 0xFF);
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_2, 0xB0, (idx >> 8) & 0x7F, idx & 0xFF);
 	apdu.le = count;
 	apdu.resplen = count;
-	apdu.resp = recvbuf;
+	apdu.resp = buf;
 
+	fixup_transceive_length(card, &apdu);
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(ctx, r, "APDU transmit failed");
 	if (apdu.resplen == 0)
-		SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_VERBOSE, sc_check_sw(card, apdu.sw1, apdu.sw2));
-	memcpy(buf, recvbuf, apdu.resplen);
+		LOG_FUNC_RETURN(ctx, sc_check_sw(card, apdu.sw1, apdu.sw2));
 
 	r =  sc_check_sw(card, apdu.sw1, apdu.sw2);
 	if (r == SC_ERROR_FILE_END_REACHED)
-		SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_VERBOSE, apdu.resplen);
-	SC_TEST_RET(ctx, SC_LOG_DEBUG_NORMAL, r, "Check SW error");
+		LOG_FUNC_RETURN(ctx, apdu.resplen);
+	LOG_TEST_RET(ctx, r, "Check SW error");
 
 	if (apdu.resplen < count)   {
 		r = iso7816_read_binary(card, idx + apdu.resplen, buf + apdu.resplen, count - apdu.resplen, flags);
 		/* Ignore all but 'corrupted data' errors */
 		if (r == SC_ERROR_CORRUPTED_DATA)
-			SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_CORRUPTED_DATA);
+			LOG_FUNC_RETURN(ctx, SC_ERROR_CORRUPTED_DATA);
 		else if (r > 0)
 			apdu.resplen += r;
 	}
 
-	SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_VERBOSE, apdu.resplen);
+	LOG_FUNC_RETURN(ctx, apdu.resplen);
 }
 
-static int iso7816_read_record(sc_card_t *card,
-			       unsigned int rec_nr, u8 *buf, size_t count,
-			       unsigned long flags)
+
+static int
+iso7816_read_record(struct sc_card *card,
+		unsigned int rec_nr, u8 *buf, size_t count, unsigned long flags)
 {
-	sc_apdu_t apdu;
-	u8 recvbuf[SC_MAX_APDU_BUFFER_SIZE];
+	struct sc_apdu apdu;
 	int r;
 
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xB2, rec_nr, 0);
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_2, 0xB2, rec_nr, 0);
 	apdu.p2 = (flags & SC_RECORD_EF_ID_MASK) << 3;
 	if (flags & SC_RECORD_BY_REC_NR)
 		apdu.p2 |= 0x04;
-	
+
 	apdu.le = count;
 	apdu.resplen = count;
-	apdu.resp = recvbuf;
+	apdu.resp = buf;
 
+	fixup_transceive_length(card, &apdu);
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if (apdu.resplen == 0)
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, sc_check_sw(card, apdu.sw1, apdu.sw2));
-	memcpy(buf, recvbuf, apdu.resplen);
+		LOG_FUNC_RETURN(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2));
 
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, apdu.resplen);
+	LOG_FUNC_RETURN(card->ctx, apdu.resplen);
 }
 
-static int iso7816_write_record(sc_card_t *card, unsigned int rec_nr,
-			        const u8 *buf, size_t count,
-			        unsigned long flags)
+
+static int
+iso7816_write_record(struct sc_card *card, unsigned int rec_nr,
+		const u8 *buf, size_t count, unsigned long flags)
 {
-	sc_apdu_t apdu;
+	struct sc_apdu apdu;
 	int r;
 
-	if (count > 256) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Trying to send too many bytes");
-		return SC_ERROR_INVALID_ARGUMENTS;
-	}
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xD2, rec_nr, 0);
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_3, 0xD2, rec_nr, 0);
 	apdu.p2 = (flags & SC_RECORD_EF_ID_MASK) << 3;
 	if (flags & SC_RECORD_BY_REC_NR)
 		apdu.p2 |= 0x04;
-	
+
 	apdu.lc = count;
 	apdu.datalen = count;
 	apdu.data = buf;
 
+	fixup_transceive_length(card, &apdu);
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card, apdu.sw1, apdu.sw2),
-		    "Card returned error");
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, count);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "Card returned error");
+
+	LOG_FUNC_RETURN(card->ctx, count);
 }
 
-static int iso7816_append_record(sc_card_t *card,
-				 const u8 *buf, size_t count,
-				 unsigned long flags)
+
+static int
+iso7816_append_record(struct sc_card *card,
+		const u8 *buf, size_t count, unsigned long flags)
 {
-	sc_apdu_t apdu;
+	struct sc_apdu apdu;
 	int r;
 
-	if (count > 256) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Trying to send too many bytes");
-		return SC_ERROR_INVALID_ARGUMENTS;
-	}
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xE2, 0, 0);
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_3, 0xE2, 0, 0);
 	apdu.p2 = (flags & SC_RECORD_EF_ID_MASK) << 3;
-	
+
 	apdu.lc = count;
 	apdu.datalen = count;
 	apdu.data = buf;
 
+	fixup_transceive_length(card, &apdu);
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card, apdu.sw1, apdu.sw2),
-		    "Card returned error");
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, count);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "Card returned error");
+
+	LOG_FUNC_RETURN(card->ctx, count);
 }
 
-static int iso7816_update_record(sc_card_t *card, unsigned int rec_nr,
-				 const u8 *buf, size_t count,
-				 unsigned long flags)
+
+static int
+iso7816_update_record(struct sc_card *card, unsigned int rec_nr,
+		const u8 *buf, size_t count, unsigned long flags)
 {
-	sc_apdu_t apdu;
+	struct sc_apdu apdu;
 	int r;
 
-	if (count > 256) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Trying to send too many bytes");
-		return SC_ERROR_INVALID_ARGUMENTS;
-	}
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xDC, rec_nr, 0);
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_3, 0xDC, rec_nr, 0);
 	apdu.p2 = (flags & SC_RECORD_EF_ID_MASK) << 3;
 	if (flags & SC_RECORD_BY_REC_NR)
 		apdu.p2 |= 0x04;
-	
+
 	apdu.lc = count;
 	apdu.datalen = count;
 	apdu.data = buf;
 
+	fixup_transceive_length(card, &apdu);
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card, apdu.sw1, apdu.sw2),
-		    "Card returned error");
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, count);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "Card returned error");
+
+	LOG_FUNC_RETURN(card->ctx, count);
 }
 
-static int iso7816_write_binary(sc_card_t *card,
-				unsigned int idx, const u8 *buf,
-				size_t count, unsigned long flags)
-{
-	sc_apdu_t apdu;
-	int r;
 
-	assert(count <= (card->max_send_size > 0 ? card->max_send_size : 255));
+static int
+iso7816_write_binary(struct sc_card *card,
+		unsigned int idx, const u8 *buf, size_t count, unsigned long flags)
+{
+	struct sc_apdu apdu;
+	int r;
 
 	if (idx > 0x7fff) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "invalid EF offset: 0x%X > 0x7FFF", idx);
+		sc_log(card->ctx, "invalid EF offset: 0x%X > 0x7FFF", idx);
 		return SC_ERROR_OFFSET_TOO_LARGE;
 	}
 
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xD0,
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_3, 0xD0,
 		       (idx >> 8) & 0x7F, idx & 0xFF);
 	apdu.lc = count;
 	apdu.datalen = count;
 	apdu.data = buf;
 
+	fixup_transceive_length(card, &apdu);
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card, apdu.sw1, apdu.sw2),
-		    "Card returned error");
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, count);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "Card returned error");
+
+	LOG_FUNC_RETURN(card->ctx, count);
 }
 
-static int iso7816_update_binary(sc_card_t *card,
-				 unsigned int idx, const u8 *buf,
-				size_t count, unsigned long flags)
+
+static int
+iso7816_update_binary(struct sc_card *card,
+		unsigned int idx, const u8 *buf, size_t count, unsigned long flags)
 {
-	sc_apdu_t apdu;
+	struct sc_apdu apdu;
 	int r;
 
-	assert(count <= (card->max_send_size > 0 ? card->max_send_size : 255));
-
 	if (idx > 0x7fff) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "invalid EF offset: 0x%X > 0x7FFF", idx);
+		sc_log(card->ctx, "invalid EF offset: 0x%X > 0x7FFF", idx);
 		return SC_ERROR_OFFSET_TOO_LARGE;
 	}
 
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xD6,
-		       (idx >> 8) & 0x7F, idx & 0xFF);
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_3, 0xD6, (idx >> 8) & 0x7F, idx & 0xFF);
 	apdu.lc = count;
 	apdu.datalen = count;
 	apdu.data = buf;
 
+	fixup_transceive_length(card, &apdu);
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, sc_check_sw(card, apdu.sw1, apdu.sw2),
-		    "Card returned error");
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, count);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "Card returned error");
+
+	LOG_FUNC_RETURN(card->ctx, count);
 }
 
 
@@ -316,6 +332,7 @@ iso7816_process_fci(struct sc_card *card, struct sc_file *file,
 {
 	struct sc_context *ctx = card->ctx;
 	size_t taglen, len = buflen;
+	int i;
 	const unsigned char *tag = NULL, *p = buf;
 
 	sc_log(ctx, "processing FCI bytes");
@@ -324,21 +341,28 @@ iso7816_process_fci(struct sc_card *card, struct sc_file *file,
 		file->id = (tag[0] << 8) | tag[1];
 		sc_log(ctx, "  file identifier: 0x%02X%02X", tag[0], tag[1]);
 	}
-	tag = sc_asn1_find_tag(ctx, p, len, 0x80, &taglen);
-	if (tag != NULL && taglen > 0 && taglen < 3) {
-		file->size = tag[0];
-		if (taglen == 2)
-			file->size = (file->size << 8) + tag[1];
+
+	/* determine the file size */
+	/* try the tag 0x80 then the tag 0x81 */
+	file->size = 0;
+	for (i = 0x80; i <= 0x81; i++) {
+		int size = 0;
+		len = buflen;
+		tag = sc_asn1_find_tag(ctx, p, len, i, &taglen);
+		if (tag == NULL)
+			continue;
+		if (taglen == 0)
+			continue;
+		if (sc_asn1_decode_integer(tag, taglen, &size) < 0)
+			continue;
+		if (size <0)
+			continue;
+
+		file->size = size;
 		sc_log(ctx, "  bytes in file: %d", file->size);
+		break;
 	}
-	if (tag == NULL) {
-		tag = sc_asn1_find_tag(ctx, p, len, 0x81, &taglen);
-		if (tag != NULL && taglen >= 2) {
-			int bytes = (tag[0] << 8) + tag[1];
-			sc_log(ctx, "  bytes in file: %d", bytes);
-			file->size = bytes;
-		}
-	}
+
 	tag = sc_asn1_find_tag(ctx, p, len, 0x82, &taglen);
 	if (tag != NULL) {
 		if (taglen > 0) {
@@ -369,9 +393,11 @@ iso7816_process_fci(struct sc_card *card, struct sc_file *file,
 			sc_log(ctx, "  EF structure: %d", byte & 0x07);
 		}
 	}
+
 	tag = sc_asn1_find_tag(ctx, p, len, 0x84, &taglen);
 	if (tag != NULL && taglen > 0 && taglen <= 16) {
 		char tbuf[128];
+
 		memcpy(file->name, tag, taglen);
 		file->namelen = taglen;
 
@@ -395,6 +421,10 @@ iso7816_process_fci(struct sc_card *card, struct sc_file *file,
 	if (tag != NULL && taglen)
 		sc_file_set_sec_attr(file, tag, taglen);
 
+	tag = sc_asn1_find_tag(ctx, p, len, 0x88, &taglen);
+	if (tag != NULL && taglen == 1)
+		file->sid = *tag;
+
 	tag = sc_asn1_find_tag(ctx, p, len, 0x8A, &taglen);
 	if (tag != NULL && taglen==1) {
 		if (tag[0] == 0x01)
@@ -404,8 +434,8 @@ iso7816_process_fci(struct sc_card *card, struct sc_file *file,
 		else if (tag[0] == 0x06 || tag[0] == 0x04)
 			file->status = SC_FILE_STATUS_INVALIDATED;
 	}
-	file->magic = SC_FILE_MAGIC;
 
+	file->magic = SC_FILE_MAGIC;
 	return SC_SUCCESS;
 }
 
@@ -417,17 +447,43 @@ iso7816_select_file(struct sc_card *card, const struct sc_path *in_path, struct
 	struct sc_apdu apdu;
 	unsigned char buf[SC_MAX_APDU_BUFFER_SIZE];
 	unsigned char pathbuf[SC_MAX_PATH_SIZE], *path = pathbuf;
-	int r, pathlen;
+	int r, pathlen, pathtype;
+	int select_mf = 0;
 	struct sc_file *file = NULL;
 
 	assert(card != NULL && in_path != NULL);
 	ctx = card->ctx;
 	memcpy(path, in_path->value, in_path->len);
 	pathlen = in_path->len;
+	pathtype = in_path->type;
+
+	if (in_path->aid.len) {
+		if (!pathlen) {
+			memcpy(path, in_path->aid.value, in_path->aid.len);
+			pathlen = in_path->aid.len;
+			pathtype = SC_PATH_TYPE_DF_NAME;
+		} else {
+			/* First, select the application */
+			sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xA4, 4, 0);
+			apdu.data = in_path->aid.value;
+			apdu.datalen = in_path->aid.len;
+			apdu.lc = in_path->aid.len;
+
+			r = sc_transmit_apdu(card, &apdu);
+			LOG_TEST_RET(ctx, r, "APDU transmit failed");
+			r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+			if (r)
+				LOG_FUNC_RETURN(ctx, r);
+
+			if (pathtype == SC_PATH_TYPE_PATH
+					|| pathtype == SC_PATH_TYPE_DF_NAME)
+				pathtype = SC_PATH_TYPE_FROM_CURRENT;
+		}
+	}
 
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0xA4, 0, 0);
 
-	switch (in_path->type) {
+	switch (pathtype) {
 	case SC_PATH_TYPE_FILE_ID:
 		apdu.p1 = 0;
 		if (pathlen != 2)
@@ -440,6 +496,7 @@ iso7816_select_file(struct sc_card *card, const struct sc_path *in_path, struct
 		apdu.p1 = 8;
 		if (pathlen >= 2 && memcmp(path, "\x3F\x00", 2) == 0) {
 			if (pathlen == 2) {	/* only 3F00 supplied */
+				select_mf = 1;
 				apdu.p1 = 0;
 				break;
 			}
@@ -466,7 +523,7 @@ iso7816_select_file(struct sc_card *card, const struct sc_path *in_path, struct
 		apdu.p2 = 0;		/* first record, return FCI */
 		apdu.resp = buf;
 		apdu.resplen = sizeof(buf);
-		apdu.le = card->max_recv_size > 0 ? card->max_recv_size : 256;
+		apdu.le = sc_get_max_recv_size(card) < 256 ? sc_get_max_recv_size(card) : 256;
 	}
 	else {
 		apdu.p2 = 0x0C;		/* first record, return nothing */
@@ -492,6 +549,19 @@ iso7816_select_file(struct sc_card *card, const struct sc_path *in_path, struct
 	if (r)
 		LOG_FUNC_RETURN(ctx, r);
 
+	if (file_out && (apdu.resplen == 0))   {
+		/* For some cards 'SELECT' MF or DF_NAME do not return FCI. */
+		if (select_mf || pathtype == SC_PATH_TYPE_DF_NAME)   {
+			file = sc_file_new();
+			if (file == NULL)
+				LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+			file->path = *in_path;
+
+			*file_out = file;
+			LOG_FUNC_RETURN(ctx, SC_SUCCESS);
+		}
+	}
+
 	if (apdu.resplen < 2)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 	switch (apdu.resp[0]) {
@@ -499,7 +569,7 @@ iso7816_select_file(struct sc_card *card, const struct sc_path *in_path, struct
 	case ISO7816_TAG_FCP:
 		file = sc_file_new();
 		if (file == NULL)
-			LOG_FUNC_RETURN(ctx,SC_ERROR_OUT_OF_MEMORY);
+			LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 		file->path = *in_path;
 		if (card->ops->process_fci == NULL) {
 			sc_file_free(file);
@@ -514,40 +584,44 @@ iso7816_select_file(struct sc_card *card, const struct sc_path *in_path, struct
 	default:
 		LOG_FUNC_RETURN(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 	}
+
 	return SC_SUCCESS;
 }
 
-static int iso7816_get_challenge(sc_card_t *card, u8 *rnd, size_t len)
+
+static int
+iso7816_get_challenge(struct sc_card *card, u8 *rnd, size_t len)
 {
 	int r;
-	sc_apdu_t apdu;
+	struct sc_apdu apdu;
 	u8 buf[10];
 
 	if (!rnd && len)
 		return SC_ERROR_INVALID_ARGUMENTS;
 
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT,
-		       0x84, 0x00, 0x00);
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0x84, 0x00, 0x00);
 	apdu.le = 8;
 	apdu.resp = buf;
 	apdu.resplen = 8;	/* include SW's */
 
 	while (len > 0) {
 		size_t n = len > 8 ? 8 : len;
-		
+
 		r = sc_transmit_apdu(card, &apdu);
-		SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+		LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 		if (apdu.resplen != 8)
 			return sc_check_sw(card, apdu.sw1, apdu.sw2);
 		memcpy(rnd, apdu.resp, n);
 		len -= n;
 		rnd += n;
-	}	
+	}
 	return 0;
 }
 
-static int iso7816_construct_fci(sc_card_t *card, const sc_file_t *file,
-	u8 *out, size_t *outlen)
+
+static int
+iso7816_construct_fci(struct sc_card *card, const sc_file_t *file,
+		u8 *out, size_t *outlen)
 {
 	u8 *p = out;
 	u8 buf[64];
@@ -556,7 +630,7 @@ static int iso7816_construct_fci(sc_card_t *card, const sc_file_t *file,
 		return SC_ERROR_BUFFER_TOO_SMALL;
 	*p++ = 0x6F;
 	p++;
-	
+
 	buf[0] = (file->size >> 8) & 0xFF;
 	buf[1] = file->size & 0xFF;
 	sc_asn1_put_tag(0x81, buf, 2, p, *outlen - (p - out), &p);
@@ -571,6 +645,7 @@ static int iso7816_construct_fci(sc_card_t *card, const sc_file_t *file,
 		switch (file->type) {
 		case SC_FILE_TYPE_INTERNAL_EF:
 			buf[0] |= 0x08;
+			/* fall through */
 		case SC_FILE_TYPE_WORKING_EF:
 			buf[0] |= file->ef_structure & 7;
 			break;
@@ -593,7 +668,7 @@ static int iso7816_construct_fci(sc_card_t *card, const sc_file_t *file,
 				p, *outlen - (p - out), &p);
 	}
 	if (file->sec_attr_len) {
-		assert(sizeof(buf) >= file->prop_attr_len);
+		assert(sizeof(buf) >= file->sec_attr_len);
 		memcpy(buf, file->sec_attr, file->sec_attr_len);
 		sc_asn1_put_tag(0x86, buf, file->sec_attr_len,
 				p, *outlen - (p - out), &p);
@@ -604,43 +679,50 @@ static int iso7816_construct_fci(sc_card_t *card, const sc_file_t *file,
 	return 0;
 }
 
-static int iso7816_create_file(sc_card_t *card, sc_file_t *file)
+
+static int
+iso7816_create_file(struct sc_card *card, sc_file_t *file)
 {
 	int r;
 	size_t len;
 	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
-	sc_apdu_t apdu;
+	struct sc_apdu apdu;
 
 	len = SC_MAX_APDU_BUFFER_SIZE;
 
 	if (card->ops->construct_fci == NULL)
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_NOT_SUPPORTED);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->construct_fci(card, file, sbuf, &len);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "construct_fci() failed");
-	
+	LOG_TEST_RET(card->ctx, r, "construct_fci() failed");
+
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xE0, 0x00, 0x00);
 	apdu.lc = len;
 	apdu.datalen = len;
 	apdu.data = sbuf;
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	return sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "Card returned error");
+
+	return r;
 }
 
-static int iso7816_get_response(sc_card_t *card, size_t *count, u8 *buf)
+
+static int
+iso7816_get_response(struct sc_card *card, size_t *count, u8 *buf)
 {
-	sc_apdu_t apdu;
+	struct sc_apdu apdu;
 	int r;
 	size_t rlen;
 
 	/* request at most max_recv_size bytes */
-	if (card->max_recv_size > 0 && *count > card->max_recv_size)
-		rlen = card->max_recv_size;
+	if (*count > sc_get_max_recv_size(card))
+		rlen = sc_get_max_recv_size(card);
 	else
 		rlen = *count;
 
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xC0, 0x00, 0x00);
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_2, 0xC0, 0x00, 0x00);
 	apdu.le      = rlen;
 	apdu.resplen = rlen;
 	apdu.resp    = buf;
@@ -648,9 +730,9 @@ static int iso7816_get_response(sc_card_t *card, size_t *count, u8 *buf)
 	apdu.flags  |= SC_APDU_FLAGS_NO_GET_RESP;
 
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if (apdu.resplen == 0)
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, sc_check_sw(card, apdu.sw1, apdu.sw2));
+		LOG_FUNC_RETURN(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2));
 
 	*count = apdu.resplen;
 
@@ -666,15 +748,17 @@ static int iso7816_get_response(sc_card_t *card, size_t *count, u8 *buf)
 	return r;
 }
 
-static int iso7816_delete_file(sc_card_t *card, const sc_path_t *path)
+
+static int
+iso7816_delete_file(struct sc_card *card, const sc_path_t *path)
 {
 	int r;
 	u8 sbuf[2];
-	sc_apdu_t apdu;
+	struct sc_apdu apdu;
 
 	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
 	if (path->type != SC_PATH_TYPE_FILE_ID || (path->len != 0 && path->len != 2)) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "File type has to be SC_PATH_TYPE_FILE_ID");
+		sc_log(card->ctx, "File type has to be SC_PATH_TYPE_FILE_ID");
 		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_INVALID_ARGUMENTS);
 	}
 
@@ -686,19 +770,25 @@ static int iso7816_delete_file(sc_card_t *card, const sc_path_t *path)
 		apdu.datalen = 2;
 		apdu.data = sbuf;
 	}
-	else /* No file ID given: means currently selected file */
+	else   {
+		/* No file ID given: means currently selected file */
 		sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0xE4, 0x00, 0x00);
-	
+	}
+
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	return sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "Card returned error");
+
+	return r;
 }
 
-static int iso7816_set_security_env(sc_card_t *card,
-				    const sc_security_env_t *env,
-				    int se_num)
+
+static int
+iso7816_set_security_env(struct sc_card *card,
+		const struct sc_security_env *env, int se_num)
 {
-	sc_apdu_t apdu;
+	struct sc_apdu apdu;
 	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
 	u8 *p;
 	int r, locked = 0;
@@ -744,20 +834,18 @@ static int iso7816_set_security_env(sc_card_t *card,
 	apdu.data = sbuf;
 	if (se_num > 0) {
 		r = sc_lock(card);
-		SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "sc_lock() failed");
+		LOG_TEST_RET(card->ctx, r, "sc_lock() failed");
 		locked = 1;
 	}
 	if (apdu.datalen != 0) {
 		r = sc_transmit_apdu(card, &apdu);
 		if (r) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
-				"%s: APDU transmit failed", sc_strerror(r));
+			sc_log(card->ctx, "%s: APDU transmit failed", sc_strerror(r));
 			goto err;
 		}
 		r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 		if (r) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
-				"%s: Card returned error", sc_strerror(r));
+			sc_log(card->ctx, "%s: Card returned error", sc_strerror(r));
 			goto err;
 		}
 	}
@@ -766,7 +854,8 @@ static int iso7816_set_security_env(sc_card_t *card,
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x22, 0xF2, se_num);
 	r = sc_transmit_apdu(card, &apdu);
 	sc_unlock(card);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+
 	return sc_check_sw(card, apdu.sw1, apdu.sw2);
 err:
 	if (locked)
@@ -774,66 +863,76 @@ err:
 	return r;
 }
 
-static int iso7816_restore_security_env(sc_card_t *card, int se_num)
+
+static int
+iso7816_restore_security_env(struct sc_card *card, int se_num)
 {
-	sc_apdu_t apdu;
+	struct sc_apdu apdu;
 	int r;
-	
+
 	assert(card != NULL);
 
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x22, 0xF3, se_num);
+
 	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	return sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "Card returned error");
+
+	return r;
 }
 
-static int iso7816_compute_signature(sc_card_t *card,
-				     const u8 * data, size_t datalen,
-				     u8 * out, size_t outlen)
+
+static int
+iso7816_compute_signature(struct sc_card *card,
+		const u8 * data, size_t datalen,
+		u8 * out, size_t outlen)
 {
 	int r;
-	sc_apdu_t apdu;
-	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
-	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
+	struct sc_apdu apdu;
 
 	assert(card != NULL && data != NULL && out != NULL);
-	if (datalen > 255)
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_INVALID_ARGUMENTS);
+	LOG_FUNC_CALLED(card->ctx);
+	sc_log(card->ctx, "ISO7816 compute signature: in-len %i, out-len %i",
+		datalen, outlen);
 
 	/* INS: 0x2A  PERFORM SECURITY OPERATION
 	 * P1:  0x9E  Resp: Digital Signature
 	 * P2:  0x9A  Cmd: Input for Digital Signature */
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x2A, 0x9E,
-		       0x9A);
-	apdu.resp = rbuf;
-	apdu.resplen = sizeof(rbuf); /* FIXME */
-	apdu.le = 256;
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_4, 0x2A, 0x9E, 0x9A);
+	apdu.resp = out;
+	apdu.resplen = outlen;
+	apdu.le = outlen;
 
-	memcpy(sbuf, data, datalen);
-	apdu.data = sbuf;
+	apdu.data = data;
 	apdu.lc = datalen;
 	apdu.datalen = datalen;
-	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
-		size_t len = apdu.resplen > outlen ? outlen : apdu.resplen;
 
-		memcpy(out, apdu.resp, len);
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, len);
-	}
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, sc_check_sw(card, apdu.sw1, apdu.sw2));
+	fixup_transceive_length(card, &apdu);
+	r = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00)
+		LOG_FUNC_RETURN(card->ctx, apdu.resplen);
+
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "Card returned error");
+
+	LOG_FUNC_RETURN(card->ctx, r);
 }
 
-static int iso7816_decipher(sc_card_t *card,
-			    const u8 * crgram, size_t crgram_len,
-			    u8 * out, size_t outlen)
+
+static int
+iso7816_decipher(struct sc_card *card,
+		const u8 * crgram, size_t crgram_len,
+		u8 * out, size_t outlen)
 {
-	int       r;
-	sc_apdu_t apdu;
-	u8        *sbuf = NULL;
+	int r;
+	struct sc_apdu apdu;
+	u8 *sbuf = NULL;
 
 	assert(card != NULL && crgram != NULL && out != NULL);
-	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_NORMAL);
+	LOG_FUNC_CALLED(card->ctx);
+	sc_log(card->ctx, "ISO7816 decipher: in-len %i, out-len %i", crgram_len, outlen);
 
 	sbuf = malloc(crgram_len + 1);
 	if (sbuf == NULL)
@@ -845,34 +944,34 @@ static int iso7816_decipher(sc_card_t *card,
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_4, 0x2A, 0x80, 0x86);
 	apdu.resp    = out;
 	apdu.resplen = outlen;
-	/* if less than 256 bytes are expected than set Le to 0x00
-	 * to tell the card the we want everything available (note: we
-	 * always have Le <= crgram_len) */
-	apdu.le      = (outlen >= 256 && crgram_len < 256) ? 256 : outlen;
-	/* Use APDU chaining with 2048bit RSA keys if the card does not do extended APDU-s */
-	if ((crgram_len+1 > 255) && !(card->caps & SC_CARD_CAP_APDU_EXT))
-		apdu.flags |= SC_APDU_FLAGS_CHAINING;
-	
+	apdu.le      = outlen;
+
 	sbuf[0] = 0; /* padding indicator byte, 0x00 = No further indication */
 	memcpy(sbuf + 1, crgram, crgram_len);
 	apdu.data = sbuf;
 	apdu.lc = crgram_len + 1;
 	apdu.datalen = crgram_len + 1;
+
+	fixup_transceive_length(card, &apdu);
 	r = sc_transmit_apdu(card, &apdu);
 	sc_mem_clear(sbuf, crgram_len + 1);
 	free(sbuf);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+
 	if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00)
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, apdu.resplen);
+		LOG_FUNC_RETURN(card->ctx, apdu.resplen);
 	else
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, sc_check_sw(card, apdu.sw1, apdu.sw2));
+		LOG_FUNC_RETURN(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2));
 }
 
-static int iso7816_build_pin_apdu(sc_card_t *card, sc_apdu_t *apdu,
+
+static int
+iso7816_build_pin_apdu(struct sc_card *card, struct sc_apdu *apdu,
 		struct sc_pin_cmd_data *data, u8 *buf, size_t buf_len)
 {
 	int r, len = 0, pad = 0, use_pin_pad = 0, ins, p1 = 0;
-	
+	int cse = SC_APDU_CASE_3_SHORT;
+
 	switch (data->pin_type) {
 	case SC_AC_CHV:
 		break;
@@ -896,11 +995,12 @@ static int iso7816_build_pin_apdu(sc_card_t *card, sc_apdu_t *apdu,
 		break;
 	case SC_PIN_CMD_CHANGE:
 		ins = 0x24;
-		if (data->pin1.len != 0 || use_pin_pad) {
+		if (data->pin1.len != 0 || (use_pin_pad && !( data->flags & SC_PIN_CMD_IMPLICIT_CHANGE))) {
 			if ((r = sc_build_pin(buf, buf_len, &data->pin1, pad)) < 0)
 				return r;
 			len += r;
-		} else {
+		}
+		else {
 			/* implicit test */
 			p1 = 1;
 		}
@@ -908,11 +1008,18 @@ static int iso7816_build_pin_apdu(sc_card_t *card, sc_apdu_t *apdu,
 		data->pin2.offset = data->pin1.offset + len;
 		if ((r = sc_build_pin(buf+len, buf_len-len, &data->pin2, pad)) < 0)
 			return r;
+		/* Special case - where provided the old PIN on the command line
+		 * but expect the new one to be entered on the keypad.
+		 */
+		if (data->pin1.len && data->pin2.len == 0) {
+			sc_log(card->ctx, "Special case - initial pin provided - but new pin asked on keypad");
+			data->flags |= SC_PIN_CMD_IMPLICIT_CHANGE;
+		};
 		len += r;
 		break;
 	case SC_PIN_CMD_UNBLOCK:
 		ins = 0x2C;
-		if (data->pin1.len != 0 || use_pin_pad) {
+		if (data->pin1.len != 0 || (use_pin_pad && !( data->flags & SC_PIN_CMD_IMPLICIT_CHANGE))) {
 			if ((r = sc_build_pin(buf, buf_len, &data->pin1, pad)) < 0)
 				return r;
 			len += r;
@@ -929,13 +1036,16 @@ static int iso7816_build_pin_apdu(sc_card_t *card, sc_apdu_t *apdu,
 			p1 |= 0x01;
 		}
 		break;
+	case SC_PIN_CMD_GET_INFO:
+		ins = 0x20;
+		/* No data to send or to receive */
+		cse = SC_APDU_CASE_1;
+		break;
 	default:
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
-	sc_format_apdu(card, apdu, SC_APDU_CASE_3_SHORT,
-				ins, p1, data->pin_reference);
-
+	sc_format_apdu(card, apdu, cse, ins, p1, data->pin_reference);
 	apdu->lc = len;
 	apdu->datalen = len;
 	apdu->data = buf;
@@ -944,16 +1054,27 @@ static int iso7816_build_pin_apdu(sc_card_t *card, sc_apdu_t *apdu,
 	return 0;
 }
 
-static int iso7816_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
-			   int *tries_left)
+
+static int
+iso7816_pin_cmd(struct sc_card *card, struct sc_pin_cmd_data *data, int *tries_left)
 {
-	sc_apdu_t local_apdu, *apdu;
+	struct sc_apdu local_apdu, *apdu;
 	int r;
 	u8  sbuf[SC_MAX_APDU_BUFFER_SIZE];
 
 	if (tries_left)
 		*tries_left = -1;
 
+	/* Many cards do support PIN status queries, but some cards don't and
+	 * mistakenly count the command as a failed PIN attempt, so for now we
+	 * whitelist cards with this flag.  In future this may be reduced to a
+	 * blacklist, subject to testing more cards. */
+	if (data->cmd == SC_PIN_CMD_GET_INFO &&
+	    !(card->caps & SC_CARD_CAP_ISO7816_PIN_INFO)) {
+		sc_log(card->ctx, "Card does not support PIN status queries");
+		return SC_ERROR_NOT_SUPPORTED;
+	}
+
 	/* See if we've been called from another card driver, which is
 	 * passing an APDU to us (this allows to write card drivers
 	 * whose PIN functions behave "mostly like ISO" except in some
@@ -967,29 +1088,27 @@ static int iso7816_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 	}
 	apdu = data->apdu;
 
-	if (!(data->flags & SC_PIN_CMD_USE_PINPAD)) {
+	if (!(data->flags & SC_PIN_CMD_USE_PINPAD) || data->cmd == SC_PIN_CMD_GET_INFO) {
 		/* Transmit the APDU to the card */
 		r = sc_transmit_apdu(card, apdu);
 
 		/* Clear the buffer - it may contain pins */
 		sc_mem_clear(sbuf, sizeof(sbuf));
-	} else {
+	}
+	else {
 		/* Call the reader driver to collect
 		 * the PIN and pass on the APDU to the card */
 		if (data->pin1.offset == 0) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
-				"Card driver didn't set PIN offset");
+			sc_log(card->ctx, "Card driver didn't set PIN offset");
 			return SC_ERROR_INVALID_ARGUMENTS;
 		}
-		if (card->reader
-		 && card->reader->ops
-		 && card->reader->ops->perform_verify) {
+		if (card->reader && card->reader->ops && card->reader->ops->perform_verify) {
 			r = card->reader->ops->perform_verify(card->reader, data);
 			/* sw1/sw2 filled in by reader driver */
-		} else {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
-				"Card reader driver does not support "
-				"PIN entry through reader key pad");
+		}
+		else {
+			sc_log(card->ctx, "Card reader driver does not support "
+					"PIN entry through reader key pad");
 			r = SC_ERROR_NOT_SUPPORTED;
 		}
 	}
@@ -998,23 +1117,78 @@ static int iso7816_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 	if (data->apdu == &local_apdu)
 		data->apdu = NULL;
 
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	if (apdu->sw1 == 0x63) {
-		if ((apdu->sw2 & 0xF0) == 0xC0 && tries_left != NULL)
-			*tries_left = apdu->sw2 & 0x0F;
-		return SC_ERROR_PIN_CODE_INCORRECT;
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu->sw1, apdu->sw2);
+
+	if (data->cmd == SC_PIN_CMD_GET_INFO) {
+		if (r == SC_ERROR_PIN_CODE_INCORRECT) {
+			data->pin1.tries_left = apdu->sw2 & 0xF;
+			r = SC_SUCCESS;
+		} else if (r == SC_ERROR_AUTH_METHOD_BLOCKED) {
+			data->pin1.tries_left = 0;
+			r = SC_SUCCESS;
+		}
+
+		if (tries_left != NULL) {
+			*tries_left = data->pin1.tries_left;
+		}
 	}
-	return sc_check_sw(card, apdu->sw1, apdu->sw2);
+
+	return r;
 }
 
-static int no_match(sc_card_t *card)
+
+static int iso7816_get_data(struct sc_card *card, unsigned int tag,  u8 *buf, size_t len)
+{
+	int                             r, cse;
+	struct sc_apdu                  apdu;
+
+	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
+
+	if (buf && len)
+		cse = SC_APDU_CASE_2;
+	else
+		cse = SC_APDU_CASE_1;
+
+	sc_format_apdu(card, &apdu, cse, 0xCA, (tag >> 8) & 0xff, tag & 0xff);
+	apdu.le = len;
+	apdu.resp = buf;
+	apdu.resplen = len;
+	r = sc_transmit_apdu(card, &apdu);
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "GET_DATA returned error");
+
+	if (apdu.resplen > len)
+		r = SC_ERROR_WRONG_LENGTH;
+	else
+		r = apdu.resplen;
+
+	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+}
+
+
+
+static int
+iso7816_init(struct sc_card *card)
+{
+#if ENABLE_SM
+	memset(&card->sm_ctx, 0, sizeof card->sm_ctx);
+#endif
+	return SC_SUCCESS;
+}
+
+
+static int
+no_match(struct sc_card *card)
 {
 	return 0;
 }
 
 static struct sc_card_operations iso_ops = {
 	no_match,
-	NULL,			/* init   */
+	iso7816_init,	/* init   */
 	NULL,			/* finish */
 	iso7816_read_binary,
 	iso7816_write_binary,
@@ -1043,7 +1217,7 @@ static struct sc_card_operations iso_ops = {
 	iso7816_process_fci,
 	iso7816_construct_fci,
 	iso7816_pin_cmd,
-	NULL,			/* get_data */
+	iso7816_get_data,
 	NULL,			/* put_data */
 	NULL,			/* delete_record */
 	NULL			/* read_public_key */
