@@ -18,7 +18,9 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <string.h>
 
@@ -579,9 +581,9 @@ int msc_extract_key(sc_card_t *card,
 
 int msc_extract_rsa_public_key(sc_card_t *card, 
 			int keyLocation,
-			int* modLength, 
+			size_t* modLength, 
 			u8** modulus,
-			int* expLength,
+			size_t* expLength,
 			u8** exponent)
 {
 	int r;
@@ -599,7 +601,9 @@ int msc_extract_rsa_public_key(sc_card_t *card,
 	if(buffer[0] != MSC_RSA_PUBLIC) SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 	*modLength = (buffer[3] << 8) | buffer[4];
 	/* Read the modulus and the exponent length */
-	
+
+	if (*modLength + 2 > sizeof buffer)
+		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_OUT_OF_MEMORY);
 	r = msc_read_object(card, inputId, fileLocation, buffer, *modLength + 2);
 	fileLocation += *modLength + 2;
 	if(r < 0) SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
@@ -640,12 +644,12 @@ int msc_compute_crypt_init(sc_card_t *card,
 	u8 *ptr;
 	int r;
 
-	u8 outputBuffer[MSC_MAX_APDU];
+	u8 outputBuffer[MSC_MAX_APDU + 2];
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x36, keyLocation, 0x01); /* Init */
 	apdu.data = buffer;
 	apdu.datalen = dataLength + 5;
 	apdu.lc = dataLength + 5;
-	
+
 	memset(outputBuffer, 0, sizeof(outputBuffer));
 	apdu.resp = outputBuffer;
 	apdu.resplen = dataLength + 2;
@@ -657,13 +661,13 @@ int msc_compute_crypt_init(sc_card_t *card,
 	*ptr = (dataLength >> 8) & 0xFF; ptr++;
 	*ptr = dataLength & 0xFF; ptr++;
 	memcpy(ptr, initData, dataLength);
-	
+
 	r = sc_transmit_apdu(card, &apdu);
 	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
 	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
 		short receivedData = outputBuffer[0] << 8 | outputBuffer[1];
-		 *outputDataLength = receivedData;
-		*outputDataLength = 0;
+		*outputDataLength = receivedData;
+
 		assert(receivedData <= MSC_MAX_APDU);
 		memcpy(outputData, outputBuffer + 2, receivedData);
 		return 0;
@@ -679,63 +683,6 @@ int msc_compute_crypt_init(sc_card_t *card,
 	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_CARD_CMD_FAILED);
 }
 
-#if 0
-int msc_compute_crypt_process(
-			sc_card_t *card, 
-			int keyLocation,
-			const u8* inputData,
-			u8* outputData,
-			size_t dataLength,
-			size_t* outputDataLength)
-{
-	sc_apdu_t apdu;
-	u8 buffer[MSC_MAX_APDU];
-	u8 outputBuffer[MSC_MAX_APDU];
-	u8 *ptr;
-	int r;
-
-	if(dataLength > MSC_MAX_SEND - 3)
-		return SC_ERROR_INVALID_ARGUMENTS;
-
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x36, keyLocation, 0x02); /* Process */
-	
-	apdu.data = buffer;
-	apdu.datalen = dataLength + 3;
-	apdu.lc = dataLength + 3;
-/* Specs say crypt returns data all the time??? But... its not implemented that way */
-	
-	memset(outputBuffer, 0, sizeof(outputBuffer));
-	apdu.resp = outputBuffer;
-	apdu.resplen = MSC_MAX_READ;
-	apdu.le = dataLength;
-	ptr = buffer;
-	*ptr = 0x01; ptr++; /* DATA LOCATION: APDU */
-	*ptr = (dataLength >> 8) & 0xFF; ptr++;
-	*ptr = dataLength & 0xFF; ptr++;
-	memcpy(ptr, inputData, dataLength);
-	
-	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	if(apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
-		short receivedData = outputBuffer[0] << 8 | outputBuffer[1];
-		 *outputDataLength = receivedData;
-		*outputDataLength = 0;
-		assert(receivedData <= MSC_MAX_APDU);
-		memcpy(outputData, outputBuffer + 2, receivedData);
-		return 0;
-	}
-	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	if (r) {
-		if (card->ctx->debug >= 2) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "process: got strange SWs: 0x%02X 0x%02X\n",
-			     apdu.sw1, apdu.sw2);
-		}
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
-	}
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_CARD_CMD_FAILED);
-}
-#endif
-
 int msc_compute_crypt_final(
 			sc_card_t *card, 
 			int keyLocation,
@@ -746,7 +693,7 @@ int msc_compute_crypt_final(
 {
 	sc_apdu_t apdu;
 	u8 buffer[MSC_MAX_APDU];
-	u8 outputBuffer[MSC_MAX_APDU];
+	u8 outputBuffer[MSC_MAX_APDU + 2];
 	u8 *ptr;
 	int r;
 
