@@ -18,7 +18,9 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <stdio.h>
 #include <ctype.h>
@@ -47,11 +49,12 @@ const char *sc_get_version(void)
 
 int sc_hex_to_bin(const char *in, u8 *out, size_t *outlen)
 {
-	int err = 0;
-	size_t left, count = 0;
+	int err = SC_SUCCESS;
+	size_t left, count = 0, in_len;
 
 	assert(in != NULL && out != NULL && outlen != NULL);
-        left = *outlen;
+	left = *outlen;
+	in_len = strlen(in);
 
 	while (*in != '\0') {
 		int byte = 0, nybbles = 2;
@@ -74,10 +77,17 @@ int sc_hex_to_bin(const char *in, u8 *out, size_t *outlen)
 			}
 			byte |= c;
 		}
+
+		/* Detect premature end of string before byte is complete */
+		if (in_len > 1 && *in == '\0' && nybbles >= 0) {
+			err = SC_ERROR_INVALID_ARGUMENTS;
+			break;
+		}
+
 		if (*in == ':' || *in == ' ')
 			in++;
 		if (left <= 0) {
-                        err = SC_ERROR_BUFFER_TOO_SMALL;
+			err = SC_ERROR_BUFFER_TOO_SMALL;
 			break;
 		}
 		out[count++] = (u8) byte;
@@ -108,7 +118,25 @@ int sc_bin_to_hex(const u8 *in, size_t in_len, char *out, size_t out_len,
 		pos += 2;
 	}
 	*pos = '\0';
-	return 0;
+	return SC_SUCCESS;
+}
+
+/*
+ * Right trim all non-printable characters
+ */
+size_t sc_right_trim(u8 *buf, size_t len) {
+
+	long i;
+
+	for(i=len-1; i >=0; i--) {
+		if(!isprint(buf[i])) {
+			buf[i] = '\0';
+			len--;
+			continue;
+		}
+		break;
+	}
+	return len;
 }
 
 u8 *ulong2bebytes(u8 *buf, unsigned long x)
@@ -256,17 +284,18 @@ void sc_format_path(const char *str, sc_path_t *path)
 {
 	int type = SC_PATH_TYPE_PATH;
 
-	memset(path, 0, sizeof(*path));
-	if (*str == 'i' || *str == 'I') {
-		type = SC_PATH_TYPE_FILE_ID;
-		str++;
+	if (path) {
+		memset(path, 0, sizeof(*path));
+		if (*str == 'i' || *str == 'I') {
+			type = SC_PATH_TYPE_FILE_ID;
+			str++;
+		}
+		path->len = sizeof(path->value);
+		if (sc_hex_to_bin(str, path->value, &path->len) >= 0) {
+			path->type = type;
+		}
+		path->count = -1;
 	}
-	path->len = sizeof(path->value);
-	if (sc_hex_to_bin(str, path->value, &path->len) >= 0) {
-		path->type = type;
-	}
-	path->count = -1;
-	return;
 }
 
 int sc_append_path(sc_path_t *dest, const sc_path_t *src)
@@ -280,7 +309,7 @@ int sc_append_path_id(sc_path_t *dest, const u8 *id, size_t idlen)
 		return SC_ERROR_INVALID_ARGUMENTS;
 	memcpy(dest->value + dest->len, id, idlen);
 	dest->len += idlen;
-	return 0;
+	return SC_SUCCESS;
 }
 
 int sc_append_file_id(sc_path_t *dest, unsigned int fid)
@@ -398,21 +427,21 @@ int sc_file_add_acl_entry(sc_file_t *file, unsigned int operation,
 	case SC_AC_NEVER:
 		sc_file_clear_acl_entries(file, operation);
 		file->acl[operation] = (sc_acl_entry_t *) 1;
-		return 0;
+		return SC_SUCCESS;
 	case SC_AC_NONE:
 		sc_file_clear_acl_entries(file, operation);
 		file->acl[operation] = (sc_acl_entry_t *) 2;
-		return 0;
+		return SC_SUCCESS;
 	case SC_AC_UNKNOWN:
 		sc_file_clear_acl_entries(file, operation);
 		file->acl[operation] = (sc_acl_entry_t *) 3;
-		return 0;
+		return SC_SUCCESS;
 	default:
 		/* NONE and UNKNOWN get zapped when a new AC is added.
 		 * If the ACL is NEVER, additional entries will be
 		 * dropped silently. */
 		if (file->acl[operation] == (sc_acl_entry_t *) 1)
-			return 0;
+			return SC_SUCCESS;
 		if (file->acl[operation] == (sc_acl_entry_t *) 2
 		 || file->acl[operation] == (sc_acl_entry_t *) 3)
 			file->acl[operation] = NULL;
@@ -422,7 +451,7 @@ int sc_file_add_acl_entry(sc_file_t *file, unsigned int operation,
 	 * of the card's AC with OpenSC's), don't add it again. */
 	for (p = file->acl[operation]; p != NULL; p = p->next) {
 		if ((p->method == method) && (p->key_ref == key_ref))
-			return 0;
+			return SC_SUCCESS;
 	}
 
 	_new = malloc(sizeof(sc_acl_entry_t));
@@ -435,13 +464,13 @@ int sc_file_add_acl_entry(sc_file_t *file, unsigned int operation,
 	p = file->acl[operation];
 	if (p == NULL) {
 		file->acl[operation] = _new;
-		return 0;
+		return SC_SUCCESS;
 	}
 	while (p->next != NULL)
 		p = p->next;
 	p->next = _new;
 
-	return 0;
+	return SC_SUCCESS;
 }
 
 const sc_acl_entry_t * sc_file_get_acl_entry(const sc_file_t *file,
@@ -518,6 +547,8 @@ void sc_file_free(sc_file_t *file)
 		free(file->prop_attr);
 	if (file->type_attr)
 		free(file->type_attr);
+	if (file->encoded_content)
+		free(file->encoded_content);
 	free(file);
 }
 
@@ -560,6 +591,8 @@ void sc_file_dup(sc_file_t **dest, const sc_file_t *src)
 		goto err;
 	if (sc_file_set_type_attr(newf, src->type_attr, src->type_attr_len) < 0)
 		goto err;
+	if (sc_file_set_content(newf, src->encoded_content, src->encoded_content_len) < 0)
+		goto err;
 	return;
 err:
 	if (newf != NULL)
@@ -606,7 +639,7 @@ int sc_file_set_prop_attr(sc_file_t *file, const u8 *prop_attr,
 			free(file->prop_attr);
 		file->prop_attr = NULL;
 		file->prop_attr_len = 0;
-		return 0;
+		return SC_SUCCESS;
 	 }
 	tmp = (u8 *) realloc(file->prop_attr, prop_attr_len);
 	if (!tmp) {
@@ -620,7 +653,7 @@ int sc_file_set_prop_attr(sc_file_t *file, const u8 *prop_attr,
 	memcpy(file->prop_attr, prop_attr, prop_attr_len);
 	file->prop_attr_len = prop_attr_len;
 
-	return 0;
+	return SC_SUCCESS;
 }
 
 int sc_file_set_type_attr(sc_file_t *file, const u8 *type_attr,
@@ -634,7 +667,7 @@ int sc_file_set_type_attr(sc_file_t *file, const u8 *type_attr,
 			free(file->type_attr);
 		file->type_attr = NULL;
 		file->type_attr_len = 0;
-		return 0;
+		return SC_SUCCESS;
 	 }
 	tmp = (u8 *) realloc(file->type_attr, type_attr_len);
 	if (!tmp) {
@@ -648,9 +681,41 @@ int sc_file_set_type_attr(sc_file_t *file, const u8 *type_attr,
 	memcpy(file->type_attr, type_attr, type_attr_len);
 	file->type_attr_len = type_attr_len;
 
-	return 0;
+	return SC_SUCCESS;
 }
 
+
+int sc_file_set_content(sc_file_t *file, const u8 *content,
+			 size_t content_len)
+{
+	u8 *tmp;
+	assert(sc_file_valid(file));
+
+	if (content == NULL) {
+		if (file->encoded_content != NULL)
+			free(file->encoded_content);
+		file->encoded_content = NULL;
+		file->encoded_content_len = 0;
+		return SC_SUCCESS;
+	}
+
+	tmp = (u8 *) realloc(file->encoded_content, content_len);
+	if (!tmp) {
+		if (file->encoded_content)
+			free(file->encoded_content);
+		file->encoded_content = NULL;
+		file->encoded_content_len = 0;
+		return SC_ERROR_OUT_OF_MEMORY;
+	}
+
+	file->encoded_content = tmp;
+	memcpy(file->encoded_content, content, content_len);
+	file->encoded_content_len = content_len;
+
+	return SC_SUCCESS;
+}
+
+
 int sc_file_valid(const sc_file_t *file) {
 #ifndef NDEBUG
 	assert(file != NULL);
@@ -726,12 +791,12 @@ int _sc_parse_atr(sc_reader_t *reader)
 		}
 	}
 	if (atr_len <= 0)
-		return 0;
+		return SC_SUCCESS;
 	if (n_hist > atr_len)
 		n_hist = atr_len;
 	reader->atr_info.hist_bytes_len = n_hist;
 	reader->atr_info.hist_bytes = p;
-	return 0;
+	return SC_SUCCESS;
 }
 
 void *sc_mem_alloc_secure(sc_context_t *ctx, size_t len)
@@ -741,34 +806,40 @@ void *sc_mem_alloc_secure(sc_context_t *ctx, size_t len)
 
     pointer = calloc(len, sizeof(unsigned char));
     if (!pointer)
-        return NULL;
+	return NULL;
 #ifdef HAVE_SYS_MMAN_H
-    /* TODO Windows support and mprotect too */
+    /* TODO mprotect */
     /* Do not swap the memory */
     if (mlock(pointer, len) >= 0)
-        locked = 1;
+	locked = 1;
+#endif
+#ifdef _WIN32
+	/* Do not swap the memory */
+	if (VirtualLock(pointer, len) != 0)
+		locked = 1;
 #endif
     if (!locked) {
-        if (ctx->paranoid_memory) {
-            sc_do_log (ctx, 0, NULL, 0, NULL, "cannot lock memory, failing allocation because paranoid set");
-            free (pointer);
-            pointer = NULL;
-        } else {
-            sc_do_log (ctx, 0, NULL, 0, NULL, "cannot lock memory, sensitive data may be paged to disk");
-        }
+	if (ctx->flags & SC_CTX_FLAG_PARANOID_MEMORY) {
+	    sc_do_log (ctx, 0, NULL, 0, NULL, "cannot lock memory, failing allocation because paranoid set");
+	    free (pointer);
+	    pointer = NULL;
+	} else {
+	    sc_do_log (ctx, 0, NULL, 0, NULL, "cannot lock memory, sensitive data may be paged to disk");
+	}
     }
     return pointer;
 }
 
 void sc_mem_clear(void *ptr, size_t len)
 {
-#ifdef ENABLE_OPENSSL
 	/* FIXME: Bug in 1.0.0-beta series crashes with 0 length */
-	if (len > 0)
+	if (len > 0)   {
+#ifdef ENABLE_OPENSSL
 		OPENSSL_cleanse(ptr, len);
 #else
-	memset(ptr, 0, len);
+		memset(ptr, 0, len);
 #endif
+	}
 }
 
 int sc_mem_reverse(unsigned char *buf, size_t len)
@@ -785,7 +856,7 @@ int sc_mem_reverse(unsigned char *buf, size_t len)
 		*(buf + len - 1 - ii) = ch;
 	}
 
-	return 0;
+	return SC_SUCCESS;
 }
 
 static int
