@@ -107,7 +107,10 @@ static CK_RV sc_pkcs11_close_session(CK_SESSION_HANDLE hSession)
 	slot->nsessions--;
 	if (slot->nsessions == 0 && slot->login_user >= 0) {
 		slot->login_user = -1;
-		slot->card->framework->logout(slot);
+		if (sc_pkcs11_conf.atomic)
+			pop_all_login_states(slot);
+		else
+			slot->p11card->framework->logout(slot);
 	}
 
 	if (list_delete(&sessions, session) != 0)
@@ -120,17 +123,17 @@ static CK_RV sc_pkcs11_close_session(CK_SESSION_HANDLE hSession)
  * the global lock held */
 CK_RV sc_pkcs11_close_all_sessions(CK_SLOT_ID slotID)
 {
-	CK_RV rv = CKR_OK;
+	CK_RV rv = CKR_OK, error;
 	struct sc_pkcs11_session *session;
 	unsigned int i;
 	sc_log(context, "real C_CloseAllSessions(0x%lx) %d", slotID, list_size(&sessions));
 	for (i = 0; i < list_size(&sessions); i++) {
 		session = list_get_at(&sessions, i);
 		if (session->slot->id == slotID)
-			if ((rv = sc_pkcs11_close_session(session->handle)) != CKR_OK)
-				return rv;
+			if ((error = sc_pkcs11_close_session(session->handle)) != CKR_OK)
+				rv = error;
 	}
-	return CKR_OK;
+	return rv;
 }
 
 CK_RV C_CloseSession(CK_SESSION_HANDLE hSession)
@@ -166,7 +169,8 @@ CK_RV C_CloseAllSessions(CK_SLOT_ID slotID)
 
 	rv = sc_pkcs11_close_all_sessions(slotID);
 
-      out:sc_pkcs11_unlock();
+out:
+	sc_pkcs11_unlock();
 	return rv;
 }
 
@@ -260,7 +264,7 @@ CK_RV C_Login(CK_SESSION_HANDLE hSession,	/* the session's handle */
 
 	slot = session->slot;
 
-	if (!(slot->token_info.flags & CKF_USER_PIN_INITIALIZED)) {
+	if (!(slot->token_info.flags & CKF_USER_PIN_INITIALIZED) && userType == CKU_USER) {
 		rv = CKR_USER_PIN_NOT_INITIALIZED;
 		goto out;
 	}
@@ -269,10 +273,16 @@ CK_RV C_Login(CK_SESSION_HANDLE hSession,	/* the session's handle */
 	if (userType == CKU_CONTEXT_SPECIFIC) {
 		if (slot->login_user == -1) {
 			rv = CKR_OPERATION_NOT_INITIALIZED;
-			goto out;
-		} else
-		rv = slot->card->framework->login(slot, userType, pPin, ulPinLen);
-	} else {
+		}
+		else   {
+			rv = restore_login_state(slot);
+			if (rv == CKR_OK)
+				rv = slot->p11card->framework->login(slot, userType, pPin, ulPinLen);
+			rv = reset_login_state(slot, rv);
+		}
+	}
+	else {
+		sc_log(context, "C_Login() slot->login_user %li", slot->login_user);
 		if (slot->login_user >= 0) {
 			if ((CK_USER_TYPE) slot->login_user == userType)
 				rv = CKR_USER_ALREADY_LOGGED_IN;
@@ -281,12 +291,22 @@ CK_RV C_Login(CK_SESSION_HANDLE hSession,	/* the session's handle */
 			goto out;
 		}
 
-		rv = slot->card->framework->login(slot, userType, pPin, ulPinLen);
+		rv = restore_login_state(slot);
+		if (rv == CKR_OK) {
+			sc_log(context, "C_Login() userType %li", userType);
+			rv = slot->p11card->framework->login(slot, userType, pPin, ulPinLen);
+			sc_log(context, "fLogin() rv %li", rv);
+		}
 		if (rv == CKR_OK)
+			rv = push_login_state(slot, userType, pPin, ulPinLen);
+		if (rv == CKR_OK) {
 			slot->login_user = userType;
+		}
+		rv = reset_login_state(slot, rv);
 	}
 
-      out:sc_pkcs11_unlock();
+out:
+	sc_pkcs11_unlock();
 	return rv;
 }
 
@@ -312,11 +332,15 @@ CK_RV C_Logout(CK_SESSION_HANDLE hSession)
 
 	if (slot->login_user >= 0) {
 		slot->login_user = -1;
-		rv = slot->card->framework->logout(slot);
+		if (sc_pkcs11_conf.atomic)
+			pop_all_login_states(slot);
+		else
+			rv = slot->p11card->framework->logout(slot);
 	} else
 		rv = CKR_USER_NOT_LOGGED_IN;
 
-      out:sc_pkcs11_unlock();
+out:
+	sc_pkcs11_unlock();
 	return rv;
 }
 
@@ -326,6 +350,7 @@ CK_RV C_InitPIN(CK_SESSION_HANDLE hSession, CK_CHAR_PTR pPin, CK_ULONG ulPinLen)
 	struct sc_pkcs11_session *session;
 	struct sc_pkcs11_slot *slot;
 
+	sc_log(context, "C_InitPIN() called, pin '%s'", pPin ? (char *) pPin : "<null>");
 	if (pPin == NULL_PTR && ulPinLen > 0)
 		return CKR_ARGUMENTS_BAD;
 
@@ -347,13 +372,19 @@ CK_RV C_InitPIN(CK_SESSION_HANDLE hSession, CK_CHAR_PTR pPin, CK_ULONG ulPinLen)
 	slot = session->slot;
 	if (slot->login_user != CKU_SO) {
 		rv = CKR_USER_NOT_LOGGED_IN;
-	} else if (slot->card->framework->init_pin == NULL) {
+	} else if (slot->p11card->framework->init_pin == NULL) {
 		rv = CKR_FUNCTION_NOT_SUPPORTED;
 	} else {
-		rv = slot->card->framework->init_pin(slot, pPin, ulPinLen);
+		rv = restore_login_state(slot);
+		if (rv == CKR_OK) {
+			rv = slot->p11card->framework->init_pin(slot, pPin, ulPinLen);
+			sc_log(context, "C_InitPIN() init-pin result %li", rv);
+		}
+		rv = reset_login_state(slot, rv);
 	}
 
-      out:sc_pkcs11_unlock();
+out:
+	sc_pkcs11_unlock();
 	return rv;
 }
 
@@ -385,7 +416,11 @@ CK_RV C_SetPIN(CK_SESSION_HANDLE hSession,
 		goto out;
 	}
 
-	rv = slot->card->framework->change_pin(slot, pOldPin, ulOldLen, pNewPin, ulNewLen);
+	rv = restore_login_state(slot);
+	if (rv == CKR_OK)
+		rv = slot->p11card->framework->change_pin(slot, pOldPin, ulOldLen, pNewPin, ulNewLen);
+	rv = reset_login_state(slot, rv);
+
 out:
 	sc_pkcs11_unlock();
 	return rv;
