@@ -48,7 +48,7 @@ myeid_delete_object(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 
 /*
  * Get 'Initialize Applet' data
- * 	using the ACLs defined in card profile.
+ * using the ACLs defined in card profile.
  */
 static int
 myeid_get_init_applet_data(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
@@ -70,7 +70,10 @@ myeid_get_init_applet_data(struct sc_profile *profile, struct sc_pkcs15_card *p1
 	sc_file_dup(&tmp_file, profile->mf_info->file);
 	if (tmp_file == NULL)
 		LOG_TEST_RET(ctx, SC_ERROR_OUT_OF_MEMORY, "Cannot duplicate MF file");
+
 	r = sc_pkcs15init_fixup_file(profile, p15card, tmp_file);
+	if (r < 0)
+		sc_file_free(tmp_file);
 	LOG_TEST_RET(ctx, r, "MF fixup failed");
 
 	/* AC 'Create DF' and 'Create EF' */
@@ -100,6 +103,8 @@ myeid_get_init_applet_data(struct sc_profile *profile, struct sc_pkcs15_card *p1
 	if (tmp_file == NULL)
 		LOG_TEST_RET(ctx, SC_ERROR_OUT_OF_MEMORY, "Cannot duplicate Application DF file");
 	r = sc_pkcs15init_fixup_file(profile, p15card, tmp_file);
+	if (r < 0)
+		sc_file_free(tmp_file);
 	LOG_TEST_RET(ctx, r, "Application DF fixup failed");
 
 	/* AC 'Create DF' and 'Create EF' */
@@ -118,8 +123,8 @@ myeid_get_init_applet_data(struct sc_profile *profile, struct sc_pkcs15_card *p1
 	else if (entry->method == SC_AC_NEVER)
 		*(data + 6) = 0xFF; /* 'NEVER'. */
 	*(data + 7) = 0xFF;
-	sc_file_free(tmp_file);
 
+	sc_file_free(tmp_file);
 	LOG_FUNC_RETURN(p15card->card->ctx, SC_SUCCESS);
 }
 
@@ -174,10 +179,10 @@ myeid_init_card(sc_profile_t *profile,
 
         r = sc_card_ctl(p15card->card, SC_CARDCTL_GET_SERIALNR, &rbuf);
 	LOG_TEST_RET(p15card->card->ctx, r,  "Get applet info failed");
-        
+
         sc_format_path("3F00", &path);
 	r = sc_select_file(p15card->card, &path, &file);
-        
+
 	if (file)
 		sc_file_free(file);
 
@@ -189,7 +194,7 @@ myeid_init_card(sc_profile_t *profile,
  */
 static int
 myeid_create_dir(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df) {
-	struct sc_context *ctx = p15card->card->ctx;
+	struct sc_context *ctx = NULL;
 	struct sc_file *file = NULL;
 	int r = 0, ii;
 	static const char *create_dfs[] = {
@@ -209,9 +214,10 @@ myeid_create_dir(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df
 		SC_PKCS15_DODF
 	};
 
-	if (!profile || !p15card || !df)
+	if (!profile || !p15card || !p15card->card || !df)
 		return SC_ERROR_INVALID_ARGUMENTS;
 
+	ctx = p15card->card->ctx;
 	LOG_FUNC_CALLED(ctx);
 	sc_log(ctx, "id (%x)", df->id);
 
@@ -222,7 +228,10 @@ myeid_create_dir(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df
 		for (ii = 0; create_dfs[ii]; ii++) {
 			sc_log(ctx, "Create '%s'", create_dfs[ii]);
 
-			if (sc_profile_get_file(profile, create_dfs[ii], &file)) {
+			r = sc_profile_get_file(profile, create_dfs[ii], &file);
+			if (file)
+				sc_file_free(file);
+			if (r) {
 				sc_log(ctx, "Inconsistent profile: cannot find %s", create_dfs[ii]);
 				LOG_FUNC_RETURN(ctx, SC_ERROR_INCONSISTENT_PROFILE);
 			}
@@ -421,9 +430,18 @@ myeid_create_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 					"Unsupported RSA key size");
 			break;
 		case SC_PKCS15_TYPE_PRKEY_EC:
-			if (sc_card_find_ec_alg(p15card->card, keybits) == NULL)
-				LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS,
-					"Unsupported EC key size");
+			/* 
+			   Here the information about curve is not available, that's why algorithm is checked
+			   without curve OID. 	  			   
+                         */
+                    
+			if (key_info->field_length != 0)
+				keybits = key_info->field_length;
+			else 
+				key_info->field_length = keybits;
+			
+			if (sc_card_find_ec_alg(p15card->card, keybits, NULL) == NULL)
+				LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported EC key size");
 			break;
 		default:
 			LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS,
@@ -437,6 +455,9 @@ myeid_create_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 	r = myeid_new_file(profile, card, object->type, key_info->key_reference, &file);
 	LOG_TEST_RET(ctx, r, "Cannot get new MyEID private key file");
 
+	if (!file || !file->path.len)
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Cannot determine private key file");
+
 	sc_log(ctx, "Key file size %d", keybits);
 	file->size = keybits;
 
@@ -474,26 +495,30 @@ myeid_store_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 	int r, keybits = key_info->modulus_length;
 
 	LOG_FUNC_CALLED(ctx);
-	
+
 	switch (object->type) {
 		case SC_PKCS15_TYPE_PRKEY_RSA:
 			if (sc_card_find_rsa_alg(p15card->card, keybits) == NULL)
 				LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported RSA key size");
 			break;
 		case SC_PKCS15_TYPE_PRKEY_EC:
-			if (sc_card_find_ec_alg(p15card->card, keybits) == NULL)
-				LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported EC key size");
+			if (!sc_valid_oid(&prkey->u.ec.params.id))
+                                if (sc_pkcs15_fix_ec_parameters(ctx, &prkey->u.ec.params))
+                                        LOG_FUNC_RETURN(ctx, SC_ERROR_OBJECT_NOT_VALID);
+						
 			if(key_info->field_length != 0)
 				keybits = key_info->field_length;
-			else 
+			else
 				key_info->field_length = keybits;
-				
 			break;
+			
+			if (sc_card_find_ec_alg(p15card->card, keybits, &prkey->u.ec.params.id) == NULL)
+				LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported algorithm or key size");			
 		default:
 			LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Store key failed: Unsupported key type");
 			break;
 	}
-	
+
 	sc_log(ctx, "store MyEID key with ID:%s and path:%s",
 			sc_pkcs15_print_id(&key_info->id), sc_print_path(&key_info->path));
 
@@ -569,13 +594,18 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 				LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported RSA key size");
 			break;
 		case SC_PKCS15_TYPE_PRKEY_EC:
-			if (sc_card_find_ec_alg(p15card->card, keybits) == NULL)
-				LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported EC key size");
+			/* EC is supported in MyEID v > 3.5. TODO: set correct return value if older MyEID version. */
+			/* Here the information about curve is not available, that's why supported algorithm is checked
+			   without curve OID. */						
+                    
 			if(key_info->field_length != 0)
 				keybits = key_info->field_length;
-			else 
+			else
 				key_info->field_length = keybits;
-				
+			
+			if (sc_card_find_ec_alg(p15card->card, keybits, NULL) == NULL)
+				LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported EC key size");
+			
 			break;
 		default:
 			LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported key type");
@@ -635,9 +665,11 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 				LOG_TEST_RET(ctx, SC_ERROR_PKCS15INIT, "Cannot get RSA key modulus: invalid key-size");
 
 			memcpy(pubkey->u.rsa.modulus.data, raw_pubkey, pubkey->u.rsa.modulus.len);
-		} 
+		}
 		else if (object->type == SC_PKCS15_TYPE_PRKEY_EC) {
+			struct sc_ec_parameters *ecparams = (struct sc_ec_parameters *)key_info->params.data;
 
+			sc_log(ctx, "curve '%s', len %i, oid '%s'", ecparams->named_curve, ecparams->field_length, sc_dump_oid(&(ecparams->id)));
 			pubkey->algorithm = SC_ALGORITHM_EC;
 
 			r = sc_select_file(card, &file->path, NULL);
@@ -651,14 +683,27 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 			r = sc_card_ctl(card, SC_CARDCTL_MYEID_GETDATA, &data_obj);
 			LOG_TEST_RET(ctx, r, "Cannot get EC public key: 'MYEID_GETDATA' failed");
 
+			if (pubkey->u.ec.ecpointQ.value)
+				free(pubkey->u.ec.ecpointQ.value);
 			pubkey->u.ec.ecpointQ.value = malloc(data_obj.DataLen - 2);
-			pubkey->u.ec.ecpointQ.len = data_obj.DataLen - 2;
-			pubkey->data.value = malloc(data_obj.DataLen);
-			pubkey->data.len = data_obj.DataLen;
-			pubkey->u.ec.params.field_length = keybits;
-			/* Omit the first 2 bytes (0x86??) */
+                        if (pubkey->u.ec.ecpointQ.value == NULL)
+				LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 			memcpy(pubkey->u.ec.ecpointQ.value, data_obj.Data + 2, data_obj.DataLen - 2);
-			memcpy(pubkey->data.value, data_obj.Data, data_obj.DataLen);
+			pubkey->u.ec.ecpointQ.len = data_obj.DataLen - 2;
+
+			if (pubkey->u.ec.params.named_curve)
+				free(pubkey->u.ec.params.named_curve);
+			pubkey->u.ec.params.named_curve = NULL;
+			if (pubkey->u.ec.params.der.value)
+				free(pubkey->u.ec.params.der.value);
+			pubkey->u.ec.params.der.value = NULL;
+			pubkey->u.ec.params.der.len = 0;
+
+			pubkey->u.ec.params.named_curve = strdup(ecparams->named_curve);
+			if (!pubkey->u.ec.params.named_curve)
+				LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+			r = sc_pkcs15_fix_ec_parameters(ctx, &pubkey->u.ec.params);
+			LOG_TEST_RET(ctx, r, "Cannot fix EC parameters");
 		}
 	}
 
