@@ -28,6 +28,7 @@
 
 #include "libopensc/pkcs15.h"
 #include "common/compat_strlcpy.h"
+#include "common/compat_strlcat.h"
 #include "util.h"
 
 static const char *app_name = "cryptoflex-tool";
@@ -145,7 +146,7 @@ static int select_app_df(void)
 
 	strcpy(str, "3F00");
 	if (opt_appdf != NULL)
-		strcat(str, opt_appdf);
+		strlcat(str, opt_appdf, sizeof str);
 	sc_format_path(str, &path);
 	r = sc_select_file(card, &path, &file);
 	if (r) {
@@ -600,8 +601,10 @@ static int create_key_files(void)
 	sc_file_add_acl_entry(file, SC_AC_OP_INVALIDATE, SC_AC_CHV, 1);
 	sc_file_add_acl_entry(file, SC_AC_OP_REHABILITATE, SC_AC_CHV, 1);
 
-	if (select_app_df())
+	if (select_app_df()) {
+		sc_file_free(file);
 		return 1;
+	}
 	r = sc_create_file(card, file);
 	sc_file_free(file);
 	if (r) {
@@ -641,7 +644,7 @@ static int read_rsa_privkey(RSA **rsa_out)
 
 static int encode_private_key(RSA *rsa, u8 *key, size_t *keysize)
 {
-	u8 buf[512], *p = buf;
+	u8 buf[1024], *p = buf;
 	u8 bnbuf[256];
 	int base = 0;
 	int r;
@@ -715,7 +718,7 @@ static int encode_private_key(RSA *rsa, u8 *key, size_t *keysize)
 
 static int encode_public_key(RSA *rsa, u8 *key, size_t *keysize)
 {
-	u8 buf[512], *p = buf;
+	u8 buf[1024], *p = buf;
 	u8 bnbuf[256];
 	int base = 0;
 	int r;
@@ -945,7 +948,7 @@ static int create_pin(void)
 	}
 	strcpy(buf, "3F00");
 	if (opt_appdf != NULL)
-		strcat(buf, opt_appdf);
+		strlcat(buf, opt_appdf, sizeof buf);
 	sc_format_path(buf, &path);
 
 	return create_pin_file(&path, opt_pin_num, "");
