@@ -50,17 +50,19 @@ static int	verbose = 0;
 
 enum {
 	OPT_SERIAL = 0x100,
-	OPT_LIST_ALG
+	OPT_LIST_ALG,
+	OPT_VERSION
 };
 
 static const struct option options[] = {
+	{ "version",		0, NULL,	OPT_VERSION },
 	{ "info",		0, NULL,		'i' },
 	{ "atr",		0, NULL,		'a' },
 	{ "serial",		0, NULL,	OPT_SERIAL  },
 	{ "name",		0, NULL,		'n' },
 	{ "get-conf-entry",	1, NULL,		'G' },
 	{ "set-conf-entry",	1, NULL,		'S' },
-	{ "list-readers",	0, NULL, 		'l' },
+	{ "list-readers",	0, NULL,		'l' },
 	{ "list-drivers",	0, NULL,		'D' },
 	{ "list-files",		0, NULL,		'f' },
 	{ "send-apdu",		1, NULL,		's' },
@@ -73,6 +75,7 @@ static const struct option options[] = {
 };
 
 static const char *option_help[] = {
+	"Prints OpenSC package revision",
 	"Prints information about OpenSC",
 	"Prints the ATR bytes of the card",
 	"Prints the card serial number",
@@ -149,7 +152,7 @@ static int opensc_get_conf_entry(const char *config)
 	}
 
 	section = buffer;
-	name = section == NULL ? NULL : strchr(section+1, ':');
+	name = strchr(section+1, ':');
 	key = name == NULL ? NULL : strchr(name+1, ':');
 	if (key == NULL) {
 		r = EINVAL;
@@ -203,7 +206,7 @@ static int opensc_set_conf_entry(const char *config)
 	}
 
 	section = buffer;
-	name = section == NULL ? NULL : strchr(section+1, ':');
+	name = strchr(section+1, ':');
 	key = name == NULL ? NULL : strchr(name+1, ':');
 	value = key == NULL ? NULL : strchr(key+1, ':');
 	if (value == NULL) {
@@ -650,12 +653,14 @@ int main(int argc, char * const argv[])
 	int do_list_files = 0;
 	int do_send_apdu = 0;
 	int do_print_atr = 0;
+	int do_print_version = 0;
 	int do_print_serial = 0;
 	int do_print_name = 0;
 	int do_list_algorithms = 0;
 	int action_count = 0;
 	const char *opt_driver = NULL;
 	const char *opt_conf_entry = NULL;
+	char **p;
 	sc_context_param_t ctx_param;
 
 	setbuf(stderr, NULL);
@@ -695,8 +700,14 @@ int main(int argc, char * const argv[])
 			action_count++;
 			break;
 		case 's':
-			opt_apdus = (char **) realloc(opt_apdus,
+			p = (char **) realloc(opt_apdus,
 					(opt_apdu_count + 1) * sizeof(char *));
+			if (!p) {
+				fprintf(stderr, "Not enough memory\n");
+				err = 1;
+				goto end;
+			}
+			opt_apdus = p;
 			opt_apdus[opt_apdu_count] = optarg;
 			do_send_apdu++;
 			if (opt_apdu_count == 0)
@@ -717,6 +728,10 @@ int main(int argc, char * const argv[])
 		case 'v':
 			verbose++;
 			break;
+		case OPT_VERSION:
+			do_print_version = 1;
+			action_count++;
+			break;
 		case 'c':
 			opt_driver = optarg;
 			break;
@@ -736,6 +751,11 @@ int main(int argc, char * const argv[])
 	if (action_count == 0)
 		util_print_usage_and_die(app_name, options, option_help, NULL);
 
+	if (do_print_version)   {
+		printf("%s\n", OPENSC_SCM_REVISION);
+		action_count--;
+	}
+
 	if (do_info) {
 		opensc_info();
 		action_count--;
@@ -751,6 +771,8 @@ int main(int argc, char * const argv[])
 		return 1;
 	}
 
+	ctx->flags |= SC_CTX_FLAG_ENABLE_DEFAULT_DRIVER;
+
 	if (verbose > 1) {
 		ctx->debug = verbose;
 		sc_ctx_log_to_file(ctx, "stderr");
