@@ -12,7 +12,7 @@
  * a generic implementation; that is how PINs and keys are stored
  * on the card. These should be implemented in pkcs15-<cardname>.c
  *
- * Copyright (C) 2002, Olaf Kirch <okir@lst.de>
+ * Copyright (C) 2002, Olaf Kirch <okir@suse.de>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -138,6 +138,9 @@ enum {
 	OPT_UPDATE_LAST_UPDATE,
 	OPT_ERASE_APPLICATION,
 	OPT_IGNORE_CA_CERTIFICATES,
+	OPT_UPDATE_EXISTING,
+	OPT_MD_CONTAINER_GUID,
+	OPT_VERSION,
 
 	OPT_PIN1     = 0x10000,	/* don't touch these values */
 	OPT_PUK1     = 0x10001,
@@ -149,6 +152,7 @@ enum {
 };
 
 const struct option	options[] = {
+	{ "version",		0, NULL,			OPT_VERSION },
 	{ "erase-card",		no_argument, NULL,		'E' },
 	{ "create-pkcs15",	no_argument, NULL,		'C' },
 	{ "store-pin",		no_argument, NULL,		'P' },
@@ -189,6 +193,7 @@ const struct option	options[] = {
 	{ "finalize",		no_argument,       NULL,	'F' },
 	{ "update-last-update", no_argument,       NULL,        OPT_UPDATE_LAST_UPDATE},
 	{ "ignore-ca-certificates",no_argument,    NULL,	OPT_IGNORE_CA_CERTIFICATES},
+	{ "update-existing",	no_argument,       NULL,	OPT_UPDATE_EXISTING},
 
 	{ "extractable",	no_argument, NULL,		OPT_EXTRACTABLE },
 	{ "insecure",		no_argument, NULL,		OPT_INSECURE },
@@ -199,6 +204,7 @@ const struct option	options[] = {
 	{ "profile",		required_argument, NULL,	'p' },
 	{ "card-profile",	required_argument, NULL,	'c' },
 	{ "options-file",	required_argument, NULL,	OPT_OPTIONS },
+	{ "md-container-guid",	required_argument, NULL,	OPT_MD_CONTAINER_GUID},
 	{ "wait",		no_argument, NULL,		'w' },
 	{ "help",		no_argument, NULL,		'h' },
 	{ "verbose",		no_argument, NULL,		'v' },
@@ -209,6 +215,7 @@ const struct option	options[] = {
 	{ NULL, 0, NULL, 0 }
 };
 static const char *		option_help[] = {
+	"Print OpenSC package version",
 	"Erase the smart card",
 	"Creates a new PKCS #15 structure",
 	"Store a new PIN/PUK on the card",
@@ -249,6 +256,7 @@ static const char *		option_help[] = {
 	"Finish initialization phase of the smart card",
 	"Update 'lastUpdate' attribut of tokenInfo",
 	"When storing PKCS#12 ignore CA certificates",
+	"Store or update existing certificate",
 
 	"Private key stored as an extractable key",
 	"Insecure mode: do not require a PIN for private key",
@@ -258,6 +266,7 @@ static const char *		option_help[] = {
 	"Specify the general profile to use",
 	"Specify the card profile to use",
 	"Read additional command line options from file",
+	"For a new key specify GUID for a MD container",
 	"Wait for card insertion",
 	"Display this message",
 	"Verbose operation. Use several times to enable debug output.",
@@ -284,6 +293,7 @@ enum {
 	ACTION_SANITY_CHECK,
 	ACTION_UPDATE_LAST_UPDATE,
 	ACTION_ERASE_APPLICATION,
+	ACTION_PRINT_VERSION,
 
 	ACTION_MAX
 };
@@ -292,6 +302,7 @@ static const char *action_names[] = {
 	"verify that card is pristine",
 	"erase card",
 	"create PKCS #15 meta structure",
+	"delete object(s)",
 	"store PIN",
 	"generate key",
 	"store private key",
@@ -300,9 +311,10 @@ static const char *action_names[] = {
 	"update certificate",
 	"store data object",
 	"finalizing card",
-	"delete object(s)",
 	"change attribute(s)",
 	"check card's sanity",
+	"update 'last-update'",
+	"erase application"
 };
 
 #define MAX_CERTS		4
@@ -345,15 +357,17 @@ static char *			opt_label = NULL;
 static char *			opt_puk_label = NULL;
 static char *			opt_pubkey_label = NULL;
 static char *			opt_cert_label = NULL;
-static char *			opt_pins[4];
+static const char *		opt_pins[4];
+static char *			pins[4];
 static char *			opt_serial = NULL;
-static char *			opt_passphrase = NULL;
+static const char *		opt_passphrase = NULL;
 static char *			opt_newkey = NULL;
 static char *			opt_outkey = NULL;
 static char *			opt_application_id = NULL;
 static char *			opt_application_name = NULL;
 static char *			opt_bind_to_aid = NULL;
 static char *			opt_puk_authid = NULL;
+static char *			opt_md_container_guid = NULL;
 static unsigned int		opt_x509_usage = 0;
 static unsigned int		opt_delete_flags = 0;
 static unsigned int		opt_type = 0;
@@ -361,6 +375,7 @@ static int			ignore_cmdline_pins = 0;
 static struct secret		opt_secrets[MAX_SECRETS];
 static unsigned int		opt_secret_count;
 static int			opt_ignore_ca_certs = 0;
+static int			opt_update_existing = 0;
 static int			verbose = 0;
 
 static struct sc_pkcs15init_callbacks callbacks = {
@@ -456,6 +471,10 @@ main(int argc, char **argv)
 		return 1;
 	}
 
+	for (n = 0; n < sizeof(pins)/sizeof(pins[0]); n++) {
+		pins[n] = NULL;
+	}
+
 	for (n = 0; n < ACTION_MAX; n++) {
 		unsigned int	action = n;
 
@@ -514,6 +533,9 @@ main(int argc, char **argv)
 			printf("About to %s.\n", action_names[action]);
 
 		switch (action) {
+		case ACTION_PRINT_VERSION:
+			printf("%s\n", OPENSC_SCM_REVISION);
+			break;
 		case ACTION_ASSERT_PRISTINE:
 			/* skip printing error message */
 			if ((r = do_assert_pristine(card)) < 0)
@@ -575,6 +597,10 @@ main(int argc, char **argv)
 		}
 	}
 
+	for (n = 0; n < sizeof(pins)/sizeof(pins[0]); n++) {
+		free(pins[n]);
+	}
+
 out:
 	if (profile) {
 		sc_pkcs15init_unbind(profile);
@@ -673,10 +699,10 @@ do_erase(sc_card_t *in_card, struct sc_profile *profile)
 		struct sc_aid aid;
 
 		aid.len = sizeof(aid.value);
-		if (sc_hex_to_bin(opt_bind_to_aid, aid.value, &aid.len))   {
+		r = sc_hex_to_bin(opt_bind_to_aid, aid.value, &aid.len);
+		if (r < 0)   {
 			fprintf(stderr, "Invalid AID value: '%s'\n", opt_bind_to_aid);
-			return 1;
-
+			goto err;
 		}
 
 		r = sc_pkcs15init_erase_card(p15card, profile, &aid);
@@ -686,6 +712,7 @@ do_erase(sc_card_t *in_card, struct sc_profile *profile)
 	}
 	ignore_cmdline_pins--;
 
+err:
 	sc_pkcs15_card_free(p15card);
 	return r;
 }
@@ -752,9 +779,10 @@ do_init_app(struct sc_profile *profile)
 
 
 	if (!opt_pins[2] && !opt_no_prompt && !opt_no_sopin) {
-		r = get_new_pin(&hints, role, "pin", &opt_pins[2]);
+		r = get_new_pin(&hints, role, "pin", &pins[2]);
 		if (r < 0)
 			goto failed;
+		opt_pins[2] = pins[2];
 	}
 
 	if (!so_puk_disabled && opt_pins[2] && !opt_pins[3] && !opt_no_prompt) {
@@ -764,9 +792,10 @@ do_init_app(struct sc_profile *profile)
 			role = "user";
 
 		hints.flags |= SC_UI_PIN_OPTIONAL;
-		r = get_new_pin(&hints, role, "puk", &opt_pins[3]);
+		r = get_new_pin(&hints, role, "puk", &pins[3]);
 		if (r < 0)
 			goto failed;
+		opt_pins[3] = pins[3];
 	}
 
 	args.so_pin = (const u8 *) opt_pins[2];
@@ -817,9 +846,11 @@ do_store_pin(struct sc_profile *profile)
 	}
 
 	sc_pkcs15init_get_pin_info(profile, SC_PKCS15INIT_USER_PIN, &info);
-	if (opt_pins[0] == NULL)
-		if ((r = get_new_pin(&hints, "user", "pin", &opt_pins[0])) < 0)
+	if (opt_pins[0] == NULL) {
+		if ((r = get_new_pin(&hints, "user", "pin", &pins[0])) < 0)
 			goto failed;
+		opt_pins[0] = pins[0];
+	}
 
 	if (*opt_pins[0] == '\0') {
 		util_error("You must specify a PIN\n");
@@ -837,9 +868,9 @@ do_store_pin(struct sc_profile *profile)
 		sc_pkcs15init_get_pin_info(profile, SC_PKCS15INIT_USER_PUK, &info);
 
 		hints.flags |= SC_UI_PIN_OPTIONAL;
-		if ((r = get_new_pin(&hints, "user", "puk", &opt_pins[1])) < 0)
+		if ((r = get_new_pin(&hints, "user", "puk", &pins[1])) < 0)
 			goto failed;
-
+		opt_pins[1] = pins[1];
 	}
 
 	if (opt_puk_authid && opt_pins[1])
@@ -995,8 +1026,7 @@ is_cacert_already_present(struct sc_pkcs15init_certargs *args)
 
 		if (!cinfo->authority)
 			continue;
-		if (args->label && objs[i]->label
-		 && strcmp(args->label, objs[i]->label))
+		if (strncmp(args->label, objs[i]->label, sizeof objs[i]->label))
 			continue;
 		/* XXX we should also match the usage field here */
 
@@ -1059,8 +1089,12 @@ do_store_certificate(struct sc_profile *profile)
 
 	memset(&args, 0, sizeof(args));
 
+	if (opt_update_existing)
+	       args.update = 1;
+
 	if (opt_objectid)
 		sc_pkcs15_format_id(opt_objectid, &args.id);
+
 	args.label = (opt_cert_label != 0 ? opt_cert_label : opt_label);
 	args.authority = opt_authority;
 
@@ -1068,8 +1102,7 @@ do_store_certificate(struct sc_profile *profile)
 	if (r >= 0)
 		r = do_convert_cert(&args.der_encoded, cert);
 	if (r >= 0)
-		r = sc_pkcs15init_store_certificate(p15card, profile,
-					&args, NULL);
+		r = sc_pkcs15init_store_certificate(p15card, profile, &args, NULL);
 
 	if (args.der_encoded.value)
 		free(args.der_encoded.value);
@@ -1200,6 +1233,10 @@ do_store_data_object(struct sc_profile *profile)
 	args.app_label = opt_application_name ? opt_application_name : "pkcs15-init";
 
 	sc_format_oid(&args.app_oid, opt_application_id);
+	if (opt_application_id && (args.app_oid.value[0] == -1))   {
+		util_error("Invalid OID \"%s\"", opt_application_id);
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	r = do_read_data_object(opt_infile, &data, &datalen);
 	if (r >= 0) {
@@ -1277,8 +1314,7 @@ static int get_cert_info(sc_pkcs15_card_t *myp15card, sc_pkcs15_object_t *certob
 	}
 
 done:
-	if (cert)
-		sc_pkcs15_free_certificate(cert);
+	sc_pkcs15_free_certificate(cert);
 	if (othercert)
 		sc_pkcs15_free_certificate(othercert);
 
@@ -1293,7 +1329,7 @@ done:
  */
 static int do_delete_crypto_objects(sc_pkcs15_card_t *myp15card,
 				struct sc_profile *profile,
-				const sc_pkcs15_id_t id,
+				const sc_pkcs15_id_t *id,
 				unsigned int which)
 {
 	sc_pkcs15_object_t *objs[10]; /* 1 priv + 1 pub + chain of at most 8 certs, should be enough */
@@ -1309,23 +1345,23 @@ static int do_delete_crypto_objects(sc_pkcs15_card_t *myp15card,
 		}
 
 		for (i = 0; i< r; i++)
-			if (sc_pkcs15_compare_id(&id, &((struct sc_pkcs15_prkey_info *)key_objs[i]->data)->id))
+			if (sc_pkcs15_compare_id(id, &((struct sc_pkcs15_prkey_info *)key_objs[i]->data)->id))
 				objs[count++] = key_objs[i];
 
 		if (!count)
-			fprintf(stderr, "NOTE: couldn't find privkey %s to delete\n", sc_pkcs15_print_id(&id));
+			fprintf(stderr, "NOTE: couldn't find privkey %s to delete\n", sc_pkcs15_print_id(id));
 	}
 
 	if (which & SC_PKCS15INIT_TYPE_PUBKEY) {
-	    if (sc_pkcs15_find_pubkey_by_id(myp15card, &id, &objs[count]) != 0)
-			fprintf(stderr, "NOTE: couldn't find pubkey %s to delete\n", sc_pkcs15_print_id(&id));
+	    if (sc_pkcs15_find_pubkey_by_id(myp15card, id, &objs[count]) != 0)
+			fprintf(stderr, "NOTE: couldn't find pubkey %s to delete\n", sc_pkcs15_print_id(id));
 		else
 			count++;
 	}
 
 	if (which & SC_PKCS15INIT_TYPE_CERT) {
-	    if (sc_pkcs15_find_cert_by_id(myp15card, &id, &objs[count]) != 0)
-			fprintf(stderr, "NOTE: couldn't find cert %s to delete\n", sc_pkcs15_print_id(&id));
+	    if (sc_pkcs15_find_cert_by_id(myp15card, id, &objs[count]) != 0)
+			fprintf(stderr, "NOTE: couldn't find cert %s to delete\n", sc_pkcs15_print_id(id));
 		else {
 			count++;
 			del_cert = 1;
@@ -1371,7 +1407,7 @@ do_delete_objects(struct sc_profile *profile, unsigned int myopt_delete_flags)
 
 	if (myopt_delete_flags & SC_PKCS15INIT_TYPE_DATA) {
 		struct sc_object_id app_oid;
-		sc_pkcs15_object_t *obj;
+		sc_pkcs15_object_t *obj = NULL;
 
 		if (opt_application_id != NULL) {
 			sc_format_oid(&app_oid, opt_application_id);
@@ -1398,7 +1434,7 @@ do_delete_objects(struct sc_profile *profile, unsigned int myopt_delete_flags)
 				util_fatal("Specify the --id for key(s) or cert(s) to be deleted\n");
 		sc_pkcs15_format_id(opt_objectid, &id);
 
-		r = do_delete_crypto_objects(p15card, profile, id, myopt_delete_flags);
+		r = do_delete_crypto_objects(p15card, profile, &id, myopt_delete_flags);
 		if (r >= 0)
 			count += r;
 	}
@@ -1475,7 +1511,7 @@ do_generate_key(struct sc_profile *profile, const char *spec)
 
 	if ((r = init_keyargs(&keygen_args.prkey_args)) < 0)
 		return r;
-        keygen_args.prkey_args.access_flags |=
+	keygen_args.prkey_args.access_flags |=
 		  SC_PKCS15_PRKEY_ACCESS_SENSITIVE
 		| SC_PKCS15_PRKEY_ACCESS_ALWAYSSENSITIVE
 		| SC_PKCS15_PRKEY_ACCESS_NEVEREXTRACTABLE
@@ -1507,7 +1543,7 @@ do_generate_key(struct sc_profile *profile, const char *spec)
 
 	if (*spec)   {
 		if (isalpha(*spec) && keygen_args.prkey_args.key.algorithm == SC_ALGORITHM_EC)   {
-			keygen_args.prkey_args.params.ec.named_curve = strdup(spec);
+			keygen_args.prkey_args.key.u.ec.params.named_curve = strdup(spec);
 			keybits = 0;
 		}
 		else {
@@ -1533,7 +1569,7 @@ static int init_keyargs(struct sc_pkcs15init_prkeyargs *args)
 		sc_pkcs15_format_id(opt_authid, &args->auth_id);
 	} else if (!opt_insecure) {
 		util_error("no PIN given for key - either use --insecure or \n"
-		      "specify a PIN using --auth-id");
+				"specify a PIN using --auth-id");
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
 	if (opt_extractable) {
@@ -1541,6 +1577,12 @@ static int init_keyargs(struct sc_pkcs15init_prkeyargs *args)
 	}
 	args->label = opt_label;
 	args->x509_usage = opt_x509_usage;
+
+	if (opt_md_container_guid)   {
+		args->guid = (unsigned char *)opt_md_container_guid;
+		args->guid_len = strlen(opt_md_container_guid);
+	}
+
 	return 0;
 }
 
@@ -1687,19 +1729,19 @@ get_pin_callback(struct sc_profile *profile,
 			switch (id) {
 			case SC_PKCS15INIT_USER_PIN:
 				name = "User PIN";
-				secret = opt_pins[OPT_PIN1 & 3];
+				secret = (char *) opt_pins[OPT_PIN1 & 3];
 				break;
 			case SC_PKCS15INIT_USER_PUK:
 				name = "User PIN unlock key";
-				secret = opt_pins[OPT_PUK1 & 3];
+				secret = (char *) opt_pins[OPT_PUK1 & 3];
 				break;
 			case SC_PKCS15INIT_SO_PIN:
 				name = "Security officer PIN";
-				secret = opt_pins[OPT_PIN2 & 3];
+				secret = (char *) opt_pins[OPT_PIN2 & 3];
 				break;
 			case SC_PKCS15INIT_SO_PUK:
 				name = "Security officer PIN unlock key";
-				secret = opt_pins[OPT_PUK2 & 3];
+				secret = (char *) opt_pins[OPT_PUK2 & 3];
 				break;
 			}
 		}
@@ -1707,22 +1749,22 @@ get_pin_callback(struct sc_profile *profile,
 			if (!(info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_SO_PIN)
 					&& !(info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_UNBLOCKING_PIN))    {
 				name = "User PIN";
-				secret = opt_pins[OPT_PIN1 & 3];
+				secret = (char *) opt_pins[OPT_PIN1 & 3];
 			}
 			else if (!(info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_SO_PIN)
 					&& (info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_UNBLOCKING_PIN))    {
 				name = "User PUK";
-				secret = opt_pins[OPT_PUK1 & 3];
+				secret = (char *) opt_pins[OPT_PUK1 & 3];
 			}
 			else if ((info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_SO_PIN)
 					&& !(info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_UNBLOCKING_PIN))    {
 				name = "Security officer PIN";
-				secret = opt_pins[OPT_PIN2 & 3];
+				secret = (char *) opt_pins[OPT_PIN2 & 3];
 			}
 			else if ((info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_SO_PIN)
 					&& (info->attrs.pin.flags & SC_PKCS15_PIN_FLAG_UNBLOCKING_PIN))    {
 				name = "Security officer PIN unlock key";
-				secret = opt_pins[OPT_PUK2 & 3];
+				secret = (char *) opt_pins[OPT_PUK2 & 3];
 			}
 		}
 		if (secret)
@@ -1780,8 +1822,11 @@ get_pin_callback(struct sc_profile *profile,
 		allocated = 1;
 	}
 
-	if (len > *pinsize)
+	if (len > *pinsize) {
+		if (allocated)
+			free(secret);
 		return SC_ERROR_BUFFER_TOO_SMALL;
+	}
 	memcpy(pinbuf, secret, len + 1);
 	*pinsize = len;
 	if (allocated)
@@ -1991,7 +2036,7 @@ do_read_private_key(const char *filename, const char *format,
 	int	r;
 
 	if (opt_passphrase)
-		passphrase = opt_passphrase;
+		passphrase = (char *) opt_passphrase;
 
 	if (!format || !strcasecmp(format, "pem")) {
 		r = do_read_pem_private_key(filename, passphrase, pk);
@@ -2020,8 +2065,12 @@ do_read_private_key(const char *filename, const char *format,
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
+	if (NULL == opt_passphrase)
+		free(passphrase);
+
 	if (r < 0)
 		util_fatal("Unable to read private key from %s\n", filename);
+
 	return r;
 }
 
@@ -2151,7 +2200,7 @@ static size_t determine_filesize(const char *filename)
 static int
 do_read_data_object(const char *name, u8 **out, size_t *outlen)
 {
-        FILE *inf;
+	FILE *inf;
 	size_t filesize = determine_filesize(name);
 	int c;
 
@@ -2440,13 +2489,13 @@ handle_option(const struct option *opt)
 		break;
 	case OPT_PIN1: case OPT_PUK1:
 	case OPT_PIN2: case OPT_PUK2:
-		opt_pins[opt->val & 3] = optarg;
+		util_get_pin(optarg, &(opt_pins[opt->val & 3]));
 		break;
 	case OPT_SERIAL:
 		opt_serial = optarg;
 		break;
 	case OPT_PASSPHRASE:
-		opt_passphrase = optarg;
+		util_get_pin(optarg, &opt_passphrase);
 		break;
 	case OPT_PUBKEY:
 		this_action = ACTION_STORE_PUBKEY;
@@ -2517,6 +2566,15 @@ handle_option(const struct option *opt)
 	case OPT_IGNORE_CA_CERTIFICATES:
 		opt_ignore_ca_certs = 1;
 		break;
+	case OPT_UPDATE_EXISTING:
+		opt_update_existing = 1;
+		break;
+	case OPT_MD_CONTAINER_GUID:
+		opt_md_container_guid = optarg;
+		break;
+	case OPT_VERSION:
+		this_action = ACTION_PRINT_VERSION;
+		break;
 	default:
 		util_print_usage_and_die(app_name, options, option_help, NULL);
 	}
@@ -2765,7 +2823,7 @@ static int verify_pin(struct sc_pkcs15_card *p15card, char *auth_id_str)
 {
 	struct sc_pkcs15_object	*pin_obj = NULL;
 	char pin_label[64];
-	char *pin;
+	char *pin = NULL;
 	int r;
 
 	if (!auth_id_str)   {
@@ -2809,7 +2867,7 @@ static int verify_pin(struct sc_pkcs15_card *p15card, char *auth_id_str)
 	}
 
 	if (opt_pins[0] != NULL)   {
-		pin = opt_pins[0];
+		pin = (char *) opt_pins[0];
 	}
 	else   {
 		sc_ui_hints_t   hints;
@@ -2817,7 +2875,7 @@ static int verify_pin(struct sc_pkcs15_card *p15card, char *auth_id_str)
                 if (opt_no_prompt)
 			return SC_ERROR_OBJECT_NOT_FOUND;
 
-		if (pin_obj->label)
+		if (pin_obj->label[0])
 			snprintf(pin_label, sizeof(pin_label), "User PIN [%s]", pin_obj->label);
 		else
 			snprintf(pin_label, sizeof(pin_label), "User PIN");
@@ -2836,6 +2894,8 @@ static int verify_pin(struct sc_pkcs15_card *p15card, char *auth_id_str)
 	if (r < 0)
 		fprintf(stderr, "Operation failed: %s\n", sc_strerror(r));
 
+	if (NULL == opt_pins[0])
+		free(pin);
+
 	return r;
 }
-
