@@ -74,6 +74,7 @@
 #define S_IWUSR S_IWRITE
 #endif
 
+#define HAVE_STRNLEN
 #define HAVE_IO_H
 #define ENABLE_PCSC
 #define HAVE_WINSCARD_H
@@ -84,15 +85,39 @@
 #define PATH_MAX FILENAME_MAX
 
 #ifndef PACKAGE_VERSION
-#define PACKAGE_VERSION "0.13.0"
+#define PACKAGE_VERSION "0.16.0"
 #endif
 
 #ifndef PACKAGE_NAME
-#define PACKAGE_NAME "opensc"
+#define PACKAGE_NAME "OpenSC"
 #endif
 
 #ifndef OPENSC_FEATURES
 #define OPENSC_FEATURES "N/A"
 #endif
 
+#ifndef OPENSC_VERSION_MAJOR
+#define OPENSC_VERSION_MAJOR 0
+#endif
+
+#ifndef OPENSC_VERSION_MINOR
+#define OPENSC_VERSION_MINOR 15
+#endif
+
+#ifndef OPENSC_VS_FF_COMPANY_NAME
+#define OPENSC_VS_FF_COMPANY_NAME "OpenSC Project"
+#endif
+
+#ifndef OPENSC_VS_FF_PRODUCT_NAME
+#define OPENSC_VS_FF_PRODUCT_NAME "OpenSC smartcard framework"
+#endif
+
+#ifndef DEFAULT_PKCS11_PROVIDER
+#define DEFAULT_PKCS11_PROVIDER "opensc-pkcs11.dll"
+#endif
+
+#ifndef OPENSC_SCM_REVISION
+#define OPENSC_SCM_REVISION "OpenSC-0.16.0, rev: 7eeba1f, commit-time: 2016-06-03 11:19:51 +0200"
+#endif
+
 #endif
